import os
import sys
import unittest
from collections import OrderedDict
from terraform_validator.ValidateUtility import ValidateUtility as class_to_test

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    """
    Prints pretty json
    :param value:
    :param htchar:
    :param lfchar:
    :param indent:
    :return: pretty json
    """


    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):
        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]

        if items:
            items = sorted(items)
        [str(item) for item in items]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)

    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)

    else:
        return repr(str(value))

class TestSecurityGroup(unittest.TestCase):
    """
    Test security group
    """

    def test_dangling_egress_rule(self):
        expected_result = [
            {
                'failure_count': '1',
                'filename': '/tf/security_group/dangling_egress_rule.tf',
                'file_results': [
                    {
                        'id': 'FATAL',
                        'type': 'VIOLATION::FAILING_VIOLATION',
                        'message': '{"Unresolved logical resource ids: [\'test\']": None}',
                        'logical_resource_ids': 'None'
                    }
                ]
            }
        ]

        if sys.version_info[0] < 3:

            expected_result = [
                {
                    'failure_count': '1',
                    'filename': '/tf/security_group/dangling_egress_rule.tf',
                    'file_results': [
                        {
                            'id': 'FATAL',
                            'type': 'VIOLATION::FAILING_VIOLATION',
                            'message': '{"Unresolved logical resource ids: [u\'test\']": None}',
                            'logical_resource_ids': 'None'
                        }
                    ]
                }
            ]

            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]

        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(
            os.path.realpath(__file__))) + '/terraform_validator/test_templates/tf/security_group/dangling_egress_rule.tf'
        debug = True

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)



    def test_security_group_missing_properties(self):

        expected_result = [
            {
                'failure_count': '1',
                'filename': '/tf/security_group/sg_missing_properties.tf',
                'file_results': [
                    {
                        'id': 'FATAL',
                        'type': 'VIOLATION::FAILING_VIOLATION',
                        'message': "{'Basic CloudFormation syntax error': [Cannot find required key 'Properties'. Path: '/Resources/sg']}",
                        'logical_resource_ids': 'None'
                    }
                ]
            }
        ]

        if sys.version_info[0] < 3:
            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(
            os.path.realpath(__file__))) + '/terraform_validator/test_templates/tf/security_group/sg_missing_properties.tf'
        debug = True

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)


    def test_security_group_when_egress_is_empty(self):
        expected_result = [
            {
                'failure_count': '1',
                'filename': '/tf/security_group/single_security_group_empty_ingress.tf',
                'file_results': [
                    {
                        'id': 'F1000',
                        'type': 'VIOLATION::FAILING_VIOLATION',
                        'message': 'Missing egress rule means all traffic is allowed outbound.  Make this explicit if it is desired configuration',
                        'logical_resource_ids': "['sg']"
                    }
                ]
            }
        ]

        if sys.version_info[0] < 3:
            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(os.path.realpath(
            __file__))) + '/terraform_validator/test_templates/tf/security_group/single_security_group_empty_ingress.tf'
        debug = False

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)

    def test_security_group_when_inline_sg_is_open_to_world(self):
        expected_result = [
            {
                'failure_count': '2',
                'filename': '/tf/security_group/two_security_group_two_cidr_ingress.tf',
                'file_results': [
                    {
                        'id': 'F1000',
                        'type': 'VIOLATION::FAILING_VIOLATION',
                        'message': 'Missing egress rule means all traffic is allowed outbound.  Make this explicit if it is desired configuration',
                        'logical_resource_ids': "['sg', 'sg2']"
                    },
                    {
                        'id': 'W2',
                        'type': 'VIOLATION::WARNING',
                        'message': 'Security Groups found with cidr open to world on ingress.  This should never be true on instance.  Permissible on ELB',
                        'logical_resource_ids': "['sg2']"
                    },
                    {
                        'id': 'W27',
                        'type': 'VIOLATION::WARNING',
                        'message': 'Security Groups found ingress with port range instead of just a single port',
                        'logical_resource_ids': "['sg', 'sg2', 'sg2']"
                    },
                    {
                        'id': 'W9',
                        'type': 'VIOLATION::WARNING',
                        'message': 'Security Groups found with ingress cidr that is not /32',
                        'logical_resource_ids': "['sg2']"
                    }
                ]
            }
        ]
        if sys.version_info[0] < 3:

            expected_result = [
                {
                    'failure_count': '2',
                    'filename': '/tf/security_group/two_security_group_two_cidr_ingress.tf',
                    'file_results': [
                        {
                            'id': 'F1000',
                            'type': 'VIOLATION::FAILING_VIOLATION',
                            'message': 'Missing egress rule means all traffic is allowed outbound.  Make this explicit if it is desired configuration',
                            'logical_resource_ids': "['sg', 'sg2']"
                        },
                        {
                            'id': 'W2',
                            'type': 'VIOLATION::WARNING',
                            'message': 'Security Groups found with cidr open to world on ingress.  This should never be true on instance.  Permissible on ELB',
                            'logical_resource_ids': "['sg2']"
                        },
                        {
                            'id': 'W27',
                            'type': 'VIOLATION::WARNING',
                            'message': 'Security Groups found ingress with port range instead of just a single port',
                            'logical_resource_ids': "['sg', 'sg2', 'sg2']"
                        },
                        {
                            'id': 'W9',
                            'type': 'VIOLATION::WARNING',
                            'message': 'Security Groups found with ingress cidr that is not /32',
                            'logical_resource_ids': "['sg2']"
                        }
                    ]
                }
            ]

            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(os.path.realpath(
            __file__))) + '/terraform_validator/test_templates/tf/security_group/two_security_group_two_cidr_ingress.tf'
        debug = False

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)

    def test_security_group_when_has_multiple_inline_egress_rules(self):
        expected_result = [
            {
                'failure_count': '0',
                'filename': '/tf/security_group/multiple_inline_egress.tf',
                'file_results': [
                ]
            }
        ]

        if sys.version_info[0] < 3:
            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(
            os.path.realpath(__file__))) + '/terraform_validator/test_templates/tf/security_group/multiple_inline_egress.tf'
        debug = False

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)

    def test_two_security_groups_ingress_standalone_with_non32_cidr(self):
        expected_result = [
            {
                'failure_count': '0',
                'filename': '/tf/security_group/non_32_cidr_standalone_ingress.tf',
                'file_results': [
                    {
                        'id': 'W9',
                        'type': 'VIOLATION::WARNING',
                        'message': 'Security Groups found with ingress cidr that is not /32',
                        'logical_resource_ids': "['securityGroupIngress2', 'securityGroupIngress3', 'securityGroupIngress4', 'securityGroupIngress5']"
                    }
                ]
            }
        ]

        if sys.version_info[0] < 3:

            expected_result = [
                {
                    'failure_count': '0',
                    'filename': '/tf/security_group/non_32_cidr_standalone_ingress.tf',
                    'file_results': [
                        {
                            'id': 'W9',
                            'type': 'VIOLATION::WARNING',
                            'message': 'Security Groups found with ingress cidr that is not /32',
                            'logical_resource_ids': "['securityGroupIngress2', 'securityGroupIngress3', 'securityGroupIngress4', 'securityGroupIngress5']"
                        }
                    ]
                }
            ]


            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(os.path.realpath(
            __file__))) + '/terraform_validator/test_templates/tf/security_group/non_32_cidr_standalone_ingress.tf'
        debug = False

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)

    def test_two_security_groups_with_non32_cidr(self):
        expected_result = [
            {
                'failure_count': '0',
                'filename': '/tf/security_group/non_32_cidr.tf',
                'file_results': [
                    {
                        'id': 'W9',
                        'type': 'VIOLATION::WARNING',
                        'message': 'Security Groups found with ingress cidr that is not /32',
                        'logical_resource_ids': "['sg', 'sg2']"
                    }
                ]
            }
        ]
        if sys.version_info[0] < 3:

            expected_result = [
                {
                    'failure_count': '0',
                    'filename': '/tf/security_group/non_32_cidr.tf',
                    'file_results': [
                        {
                            'id': 'W9',
                            'type': 'VIOLATION::WARNING',
                            'message': 'Security Groups found with ingress cidr that is not /32',
                            'logical_resource_ids': "['sg', 'sg2']"
                        }
                    ]
                }
            ]

            new_file_results = []

            for info in expected_result[0]['file_results']:
                print('info: ' + str(info))
                print('type: ' + str(type(info)))
                order_of_keys = ["id", "type", "message", "logical_resource_ids"]

                new_results = OrderedDict()
                for key in order_of_keys:
                    new_results[key] = info[key]

                new_file_results.append(new_results)
                print('new file results: ' + str(new_file_results))

                expected_result[0]['file_results'] = new_file_results

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
            expected_result = [OrderedDict(list_of_tuples)]


        expected_result = pretty(expected_result)

        template_name = os.path.dirname(os.path.dirname(
            os.path.realpath(__file__))) + '/terraform_validator/test_templates/tf/security_group/non_32_cidr.tf'
        debug = False

        config_dict = {}
        config_dict['template_file'] = template_name
        config_dict['debug'] = debug
        config_dict['profile'] = None
        config_dict['rules_directory'] = None
        config_dict['input_path'] = None
        config_dict['profile'] = None
        config_dict['allow_suppression'] = False
        config_dict['print_suppression'] = False
        config_dict['parameter_values_path'] = None
        config_dict['isolate_custom_rule_exceptions'] = None
        validator = class_to_test(config_dict)

        real_result = validator.validate()
        self.maxDiff = None
        print('expected results: ' + str(expected_result))
        print('real results: ' + str(real_result))

        self.assertEqual(expected_result, real_result)

    def test_multiple_security_groups(self):

      expected_result = [
            {
                'failure_count': '0',
                'filename': '/tf/security_group/multiple_ingress_security_groups.tf',
                'file_results': [
                    {
                        'id': 'W5',
                        'type': 'VIOLATION::WARNING',
                        'message': 'Security Groups found with cidr open to world on egress',
                        'logical_resource_ids': "['emrSecurityGroup']"
                    }
                ]
            }
        ]

      if sys.version_info[0] < 3:
          new_file_results = []

          for info in expected_result[0]['file_results']:
              print('info: ' + str(info))
              print('type: ' + str(type(info)))
              order_of_keys = ["id", "type", "message", "logical_resource_ids"]

              new_results = OrderedDict()
              for key in order_of_keys:
                  new_results[key] = info[key]

              new_file_results.append(new_results)
              print('new file results: ' + str(new_file_results))

              expected_result[0]['file_results'] = new_file_results

          order_of_keys = ["failure_count", "filename", "file_results"]
          list_of_tuples = [(key, expected_result[0][key]) for key in order_of_keys]
          expected_result = [OrderedDict(list_of_tuples)]

      expected_result = pretty(expected_result)

      template_name = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))+'/terraform_validator/test_templates/tf/security_group/multiple_ingress_security_groups.tf'
      debug = False

      config_dict = {}
      config_dict['template_file'] = template_name
      config_dict['debug'] = debug
      config_dict['profile']= None
      config_dict['rules_directory'] = None
      config_dict['input_path'] = None
      config_dict['profile'] = None
      config_dict['allow_suppression'] = False
      config_dict['print_suppression'] = False
      config_dict['parameter_values_path'] = None
      config_dict['isolate_custom_rule_exceptions'] = None
      validator = class_to_test(config_dict)

      real_result =  validator.validate()
      self.maxDiff = None

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.assertEqual(expected_result, real_result)




from __future__ import absolute_import, division, print_function
from future.types.newstr import newstr
import inspect
import re
import sys
import json
import os
from builtins import (str)
from collections import OrderedDict
from terraform_validator.Violation import Violation

def lineno():
    """Returns the current line number in our program."""
    return str(' - JsonResults - line number: '+str(inspect.currentframe().f_back.f_lineno))

class JsonResults:
    """
    Json results
    """
    def __init__(self, debug=False, suppress_errors=False):
        """
        Initialize JsonResults
        :param debug:
        """
        self.debug = debug
        self.suppress_errors = suppress_errors
        if self.debug:
            print('JsonResults - init'+lineno())



    def pretty(self, value, htchar='\t', lfchar='\n', indent=0):
        """
        Prints pretty json
        :param value:
        :param htchar:
        :param lfchar:
        :param indent:
        :return: pretty json
        """

        if self.debug:
            print('### type: '+str(type(value))+lineno()+lineno())
        if type(value) == type(OrderedDict()):
            if self.debug:
                print('found it'+lineno())
                print('value: '+str(value)+lineno())


        nlch = lfchar + htchar * (indent + 1)
        if type(value) == type(dict()) or type(value) == type(OrderedDict()):
            if (self.debug):
                print("##############################\n")
                print('is dict'+lineno())
                print("##############################\n")

            items = [
                nlch + repr(key) + ': ' + self.pretty(value[key], htchar, lfchar, indent + 1)
                for key in value
            ]
            return str('{%s}' % (','.join(items) + lfchar + htchar * indent)).replace('\'','"').replace(']"',']').replace('"[','[').replace('"{"','{"').replace('"}"','"}').replace('"[\'','[\'').replace('\']"','\']')

        elif type(value) == type(list()):
            if (self.debug):
                print("##############################\n")
                print('is list'+lineno())
                print("##############################\n")


            items = [
                nlch + self.pretty(item, htchar, lfchar, indent + 1)
                for item in value
            ]
            if self.debug:
                print('items: '+str(items)+lineno())
                print('items type: '+str(type(items))+lineno())
            if items:
                items = sorted(items)
            [str(item) for item in sorted(items)]
            return str('[%s]' % (','.join(sorted(items)) + lfchar + htchar * indent)).replace('\'','"').replace(']"',']').replace('"[','[').replace('"{"','{"').replace('"}"','"}').replace('"[\'','[\'').replace('\']"','\']')

        elif type(value) is tuple:
            if (self.debug):
                print("##############################\n")
                print('is tuple'+lineno())
                print("##############################\n")

            items = [
                nlch + self.pretty(item, htchar, lfchar, indent + 1)
                for item in sorted(value)
            ]

            if self.debug:
                print('returning: '+str('(%s)' % (','.join(items) + lfchar + htchar * indent))+lineno())
            return str('(%s)' % (','.join(items) + lfchar + htchar * indent)).replace('\'','"').replace(']"',']').replace('"[','[').replace('"{"','{"').replace('"}"','"}').replace('"[\'','[\'').replace('\']"','\']')

        elif type(value) == type(newstr()):
            if self.debug:
                print('is a new string'+lineno())
                print('value: '+str(value)+lineno())

            if value.startswith('['):
                if self.debug:
                    print('starts with bracket'+lineno())
                my_json = eval((value.replace('\'','"')))
                my_new_list = []
                for my_it in sorted(my_json):
                    my_new_list.append(str(my_it))
                return str('"'+str(my_new_list)+'"').replace(']"',']').replace('"[','[').replace('"{"','{"').replace('"}"','"}').replace('"[\'','[\'').replace('\']"','\']')
                #value = self.pretty(my_json, htchar, lfchar, indent + 1)



            return repr(str(value))

        else:
            if (self.debug):
                print("##############################\n")
                print('is other: '+str(type(value))+lineno())
                print('class name: '+str(value.__class__.__name__))
                print('returning: '+str(value))
                print("##############################\n")

            return str(repr(str(value))).replace('\'','"').replace(']"',']').replace('"[','[').replace('"{"','{"').replace('"}"','"}').replace('"[\'','[\'').replace('\']"','\']')


    def render(self, results):
        """
        Renders results
        :param results:
        :return:
        """
        if self.debug:
            print("########################################")
            print('render:'+lineno())

            print('results: '+str(results)+lineno())
            print('type: '+str(type(results))+lineno())
            print("#########################################\n")
        array_of_results = []

        # This is an array of violations
        if type(results)==type(list()):

            if self.debug:
                print('is a list: '+lineno())
                print('list count: '+str(len(results))+lineno())

            ordered_violations_by_filename = {}

            # Iterate over each of the violqtions
            for r in results:
                if self.debug:

                    print("\n"+'################################')
                    print('violation: '+str(r)+lineno())
                    print('###################################'+"\n")

                # Not doing anything
                if 'filename' in r:

                    if self.debug:
                        print('r: '+str(r)+lineno())
                        print('filename: '+str(r['filename'])+lineno())

                    matchObj = re.match(r'(.*)(/json.*)', r['filename'], re.M | re.I)

                    if matchObj:
                        r['filename'] = matchObj.group(2)
                    else:
                        if self.debug:
                            print('no match '+lineno())


                # Not doing anything
                if 'file_results' in r:
                    if self.debug:
                        print('threre are file results: '+str(r['file_results'])+lineno())

                    if 'failure_count' in r['file_results']:
                        r['file_results'].pop('failure_count', None)

                new_violations_list = []
                # If there are violations in the file results
                # Lets put the violations in to an OrderedDict
                if 'violations' in r['file_results']:

                    if self.debug:
                        print('found violations: '+lineno())
                        print('type: '+str(type(r['file_results']['violations']))+lineno())

                    # Lets first just iterate over the violations and get the id so
                    # we can sort
                    if type(r['file_results']['violations']) == type(list()):

                        # We are trying to convert each violation in the list to an ordered dictionary
                        # and put the ordered lists in order by violation id
                        if self.debug:
                            print('type is a list'+lineno())
                        new_violations = []
                        for violation in r['file_results']['violations']:

                            if type(violation) == type(list()):
                                if self.debug:
                                    print('violation is a list'+lineno())


                                temp_list = []

                                for items in violation:
                                    temp_list.append(items.to_hash())

                                new_violations.append(temp_list)

                            elif hasattr(violation, '__class__') and hasattr(violation.__class__,'__name__') and violation.__class__.__name__ == 'Violation':
                                 if self.debug:
                                     print('has class' + lineno())
                                     print('class: ' + str(violation.__class__.__name__) + lineno())

                                 new_violations.append(violation.to_hash())
                            else:

                                if self.debug:
                                    print('violation is not a list: '+lineno())

                                new_violations.append(violation.to_hash())

                        r['file_results'] = new_violations


                    elif type(r['file_results']['violations']) == type(dict()):
                        if self.debug:
                            print('is a dict'+lineno())
                        r['file_results'] = [violation.to_hash()]


                # If there are no violations in file results
                else:
                    if self.debug:
                        print('there are no violations: '+lineno())
                        print('type:'+str(type(r['file_results']))+lineno())
                    for item in r['file_results']:
                        if self.debug:
                            print('item: '+str(item)+lineno())
                            print('type: '+str(type(item))+lineno())
                        if hasattr(item,'__name__'):
                            if self.debug:
                                print('has name'+lineno())
                                print('name: '+str(item.__name__)+lineno())
                        if hasattr(item, '__class__') and hasattr(item.__class__,'__name__') and item.__class__.__name__ == 'Violation':
                            if self.debug:
                                print('has class' + lineno())
                                print('class: ' + str(item.__class__.__name__) + lineno())
                                print('dirs:'+str(dir(item))+lineno())


                                print('vars: '+str(vars(item))+lineno())

                            r['file_results'] = [item.to_hash()]


            hash = results

            order_of_keys = ["failure_count", "filename", "file_results"]


            if hash:
                if self.debug:
                    print("\n"+'## there is a hash: '+str(hash)+lineno())

                if type(hash) == type(list()):
                    if self.debug:
                        print('ist a list: '+str(lineno()))
                    # Iterate over each violation

                    if sys.version_info[0] < 3:
                        for item in sorted(hash):
                            if self.debug:
                                print("\n#################################")
                                print('item: '+str(item)+lineno())
                                print("#####################################\n")

                            my_ordered_dict= OrderedDict()
                            for key in order_of_keys:
                                my_ordered_dict[key] = item[key]

                            if self.debug:
                                print("\n"+'ordered dict is: '+str(my_ordered_dict)+lineno())

                            array_of_results.append(my_ordered_dict)

                            if self.debug:
                                print("\n"+'results: '+str(array_of_results)+lineno())
                    else:
                        for item in hash:
                            if self.debug:
                                print("\n#################################")
                                print('item: ' + str(item) + lineno())
                                print("#####################################\n")

                            my_ordered_dict = OrderedDict()
                            for key in order_of_keys:
                                my_ordered_dict[key] = item[key]

                            if self.debug:
                                print("\n" + 'ordered dict is: ' + str(my_ordered_dict) + lineno())

                            array_of_results.append(my_ordered_dict)

                            if self.debug:
                                print("\n" + 'results: ' + str(array_of_results) + lineno())

                # if not a list - FIXME
                else:
                    if self.debug:
                        print('is not a list: '+lineno())
                    for item in hash:
                        if self.debug:
                            print('item: '+str(item)+lineno())
                            print('hash item: '+str(hash[item])+lineno())

                        my_ordered_dict = OrderedDict()

                        for key in order_of_keys:
                            my_ordered_dict[key]= item[key]

                    if self.debug:
                        print('my ordered dict: '+str(my_ordered_dict)+lineno())


                    array_of_results.append(my_ordered_dict)


        # If the results is not a list, then assuming it is a hash...FIXME
        else:
            if self.debug:
                print('results is not a list: '+lineno())

            order_of_keys = ["failure_count", "filename", "file_results"]
            list_of_tuples = [(key, results[key]) for key in order_of_keys]
            new_results = OrderedDict(list_of_tuples)

            hash=[]
            for result in new_results:
                if self.debug:
                    print('result: '+str(result)+lineno())

                for item in result:

                    if item:
                        array_of_results.append(item.to_hash())

            if self.debug:
                print('returning: '+str(hash)+lineno())

        return self.pretty(array_of_results)


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' - IamPolicyWildcardResourceRule- caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


class IamPolicyWildcardResourceRule(BaseRule):

  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model:
    """
    BaseRule.__init__(self, cfn_model,debug=debug)

  def rule_text(self):
    """
    Get rule text
    :return:
    """
    if self.debug:
        print('rule_text'+lineno())
    return 'IAM policy should not allow * resource'


  def rule_type(self):
    """
    Get rule type
    :return:
    """
    self.type= 'VIOLATION::WARNING'
    return 'VIOLATION::WARNING'

  def rule_id(self):
    """
    Get rule id
    :return:
    """
    if self.debug:
        print('rule_id'+lineno())
    self.id ='W12'
    return 'W12'


  def audit_impl(self):
    """
    Audit
    :return: violations
    """
    if self.debug:
        print('IamPolicyWildcardResourceRule - audit_impl'+lineno())

    violating_policies = []

    resources = self.cfn_model.resources_by_type('AWS::IAM::Policy')

    if len(resources)>0:
      for resource in resources:
          if self.debug:
              print('resource: '+str(resource)+lineno())

          if hasattr(resource,'policy_document'):
            if self.debug:
                print('has policy document '+lineno())
                print(resource.policy_document.statements)
            if resource.policy_document.wildcard_allowed_resources():
                if self.debug:
                    print('has wildcard allowed resource')

                violating_policies.append(str(resource.logical_resource_id))
    else:
      if self.debug:
        print('no violating_policies' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_policies


from __future__ import absolute_import, division, print_function
import yaml
import sys
import os
import json
from terraform_model.validator.ResourceTypeValidator import ResourceTypeValidator
import inspect
from builtins import (str)

def lineno():
    """Returns the current line number in our program."""
    return str(' -  SchemaGenerator- line number: '+str(inspect.currentframe().f_back.f_lineno))

class SchemaGenerator:
    """
    # This generator is a bit of hacking to trick kwalify into validating yaml for a cfn template.
    #
    # because cfn uses open-ended key names for the resources.... a static schema can't be used
    # with kwalify.  so first we make sure there is a basic structure of resources with Type values
    # then we generate the schema from the document for the keys and cross-reference with schema
    # files per resource type
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug:
        """

        self.debug= debug

        if self.debug:
            print(' __init__' + lineno())

    def generate(self, cloudformation_yml):
        """
        Generate schema
        :param cloudformation_yml:
        :return:
        """

        # make sure structure of Resources is decent and that every record has a Type at least
        validator = ResourceTypeValidator(debug=self.debug)
        cloudformation_hash = validator.validate(cloudformation_yml)

        parameters_schema = self.generate_schema_for_parameter_keys(cloudformation_hash)

        if self.debug:
            print('getting resources schema: '+lineno())

        resources_schema = self.generate_schema_for_resource_keys(cloudformation_hash)

        with open(os.path.dirname(os.path.dirname(__file__)) + '/schema/schema.yml.erb', 'r') as stream:
            main_schema = yaml.load(stream)

        if not parameters_schema or len(parameters_schema)<1:
            if self.debug:
                print('not variables in variable schema'+lineno())

            main_schema.pop('mapping', 'variable')
            #main_schema['mapping'].delete 'Parameters'
        else:
            if self.debug:
                print('there are variables'+lineno())
            main_schema['mapping']['variable']['mapping'] = parameters_schema

        main_schema['mapping']['resource']['mapping'] = resources_schema

        if self.debug:
            print("\n###parameter schema: "+str(parameters_schema)+lineno())
            print("\n###main_schema"+str(main_schema)+lineno())

        return main_schema



    ##
    # this is fairly superfluous.  there's not much structure here
    # except that Types are Strings.... anything else is up to a rule
    # to wade through all the optional crap (like looking for NoEcho)
    def generate_schema_for_parameter_keys(self, cloudformation_hash):
          """
          Generate schema for parameter keys
          :param cloudformation_hash:
          :return:
          """
          if self.debug:
            print('generate_schema_for_parameter_keys '+lineno())
            print('cloudformation_hash '+str(cloudformation_hash)+lineno())

          if not cloudformation_hash:
              return {}

          if self.debug:
              print('cloudformation is a hash: '+lineno())

          parameters_schema = {
            '=' : { 'type' : 'any'}
          }

          if 'variable' in cloudformation_hash:

              if self.debug:
                  print('variable in cloudformation hash: '+lineno())

              for parameter in cloudformation_hash['variable']:
                  for parameter in cloudformation_hash['variable'][parameter]:
                      parameters_schema[parameter]= {
                      'type' : 'map',
                      'mapping' : {
                        'Type' : {
                          'type' : 'str'
                        },
                        '=' : {
                          'type' : 'any'
                        }
                      }
                    }

          if self.debug:
              print('parameters_schema: '+str(parameters_schema))

          return parameters_schema


    def generate_schema_for_resource_keys(self, cloudformation_hash):
        """
        Generate schema for resource keys
        :param cloudformation_hash:
        :return:
        """

        if type(cloudformation_hash) == type(str()):
            json_acceptable_string = cloudformation_hash.replace("'", "\"")
            cloudformation_hash= json.loads(json_acceptable_string)

        resources_schema = {
          '=' : { 'type' : 'any'}
        }

        for resources in cloudformation_hash['resource']:

            if self.debug:
                print('resource: '+str(resources)+lineno())


            schema_hash=self.schema_for_type(resources)

            if schema_hash:

                if self.debug:
                    print('schema_hash:'+str(schema_hash)+lineno())

                resources_schema['resource'][resource]= schema_hash

        if self.debug:
            print('resoures_schema: '+str(resources_schema)+lineno())

        return resources_schema


    def schema_file(self, file):
         """
         ???
         :param file:
         :return:
         """
         if self.debug:
            print('schema_file '+lineno())

         file = file.replace('::','_')+'.yml'

         if self.debug:
             print('file: '+str(file)+lineno())
             print('path: '+str(os.path.dirname(os.path.dirname(__file__))))

         path = os.path.dirname(os.path.dirname(__file__))+'/schema'

         file = str(path)+'/'+str(file)
         if os.path.exists(file):

             if self.debug:
                 print('found file: '+str(file)+lineno())
             return file
         return None


    def schema_for_type(self, type):
         """
         ???
         :param type:
         :return:
         """
         if self.debug:
            print('schema_for_type:'+lineno())
            print('type: '+str(type)+lineno())

         yaml_file = self.schema_file(type)

         if yaml_file:

             # Read YAML file
             with open(yaml_file, 'r') as stream:
                 data_loaded = yaml.load(stream)

                 if self.debug:
                     print('data: '+str(data_loaded)+lineno())

                 return data_loaded

         return None


import os
import sys
import unittest
from collections import OrderedDict
from terraform_model.model.References import References as class_to_test
from terraform_model.model.CfnModel import CfnModel

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):

        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)
    else:
        return repr(value)

class TestReferencesModel(unittest.TestCase):
    """
    Test ipaddr
    """

    def test_references(self):


      cfn_model = {
          "Parameters": {
              "test": {
                  "Type": "String",
                  "Default": "subnet-4fd01116"
              }
          },
          "Resources": {
              "myENI": {
                  "Type": "AWS::EC2::NetworkInterface",
                  "Properties": {
                      "Tags": [{"Key": "foo", "Value": "bar"}],
                      "Description": "A nice description.",
                      "SourceDestCheck": "false",
                      "GroupSet": ["sg-75zzz219"],
                      "SubnetId": "subnet-3z648z53",
                      "PrivateIpAddress": "10.0.0.16"
                  }
              }
          }
        }

      model = CfnModel()
      model.raw_model = cfn_model
      model.transform_hash_into_parameters(cfn_model)

      value={"Ref":"test"}

      expected_result = {"Ref":"test"}

      real_result = class_to_test.resolve_value(cfn_model=model, value=value, debug=True)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))
      print('type: '+str(type(real_result)))
      print('name: '+str(real_result.__class__.__name__))
      self.maxDiff = None
      self.assertEqual(expected_result, real_result)




import os
import sys
import unittest
from collections import OrderedDict
from terraform_validator.IpAddr import IpAddr as class_to_test

def pretty(value, htchar='\t', lfchar='\n', indent=0):
    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):

        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)
    else:
        return repr(value)

class TestIpAddr(unittest.TestCase):
    """
    Test ipaddr
    """

    def test_ip4_open(self):


      expected_result = True


      dict = {}
      dict['CidrIp'] = '0.0.0.0/0'

      real_result = class_to_test.ip4_open(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip4_open_list(self):


      expected_result = True


      dict = []
      dict.append({'CidrIp': '0.0.0.0/0'})

      real_result = class_to_test.ip4_open(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip4_not_open(self):


      expected_result = False


      dict = {}
      dict['CidrIp'] = '192.168.1.0/32'

      real_result = class_to_test.ip4_open(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip4_not_open_list(self):


      expected_result = False


      dict = []
      dict.append({'CidrIp':'192.168.1.0/32'})

      real_result = class_to_test.ip4_open(ingress=dict, debug=True)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)



    def test_ip6_open(self):


      expected_result = True


      dict = {}
      dict['CidrIp'] = '::/0'

      real_result = class_to_test.ip6_open(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)



    def test_ip6_open_list(self):


      expected_result = True


      dict = []
      dict.append({'CidrIp':'::/0'})

      real_result = class_to_test.ip6_open(ingress=dict, debug=True)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip6_not_open(self):


      expected_result = False


      dict = {}
      dict['CidrIp'] = '::/32'

      real_result = class_to_test.ip6_open(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip6_not_open_list(self):


      expected_result = False


      dict = []
      dict.append({'CidrIp': '::/32'})

      real_result = class_to_test.ip6_open(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)



    def test_ip4_no_range(self):


      expected_result = True


      dict = {}
      dict['CidrIp'] = '192.168.1.0/32'

      real_result = class_to_test.ip4_cidr_range(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)

    def test_ip4_range(self):


      expected_result = False


      dict = {}
      dict['CidrIp'] = '192.168.1.0/64'

      real_result = class_to_test.ip4_cidr_range(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip4_range_list(self):


      expected_result = False


      dict = []
      dict.append( {'CidrIp':'192.168.1.0/64'})

      real_result = class_to_test.ip4_cidr_range(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)



    def test_ip6_no_range(self):


      expected_result = True


      dict = {}
      dict['CidrIp'] = '2001:0db8:85a3:0000:0000:8a2e:0370/128'

      real_result = class_to_test.ip6_cidr_range(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


    def test_ip6_range(self):


      expected_result = False


      dict = {}
      dict['CidrIp'] = '2001:0db8:85a3:0000:0000:8a2e:0370/64'

      real_result = class_to_test.ip6_cidr_range(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)

    def test_ip6_range_list(self):


      expected_result = False


      dict = []
      dict.append({'CidrIp': '2001:0db8:85a3:0000:0000:8a2e:0370/64'})

      real_result = class_to_test.ip6_cidr_range(ingress=dict, debug=False)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))

      self.maxDiff = None
      self.assertEqual(expected_result, real_result)


from __future__ import absolute_import, division, print_function
import inspect
import sys


def lineno():
    """Returns the current line number in our program."""
    return str(' - Principal - line number: '+str(inspect.currentframe().f_back.f_lineno))


class Principal:
    """
    Principal model
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug: 
        """
        self.debug = debug

        if self.debug:
            print('__init__'+lineno())

    def wildcard(self, principal):
        """
        Whether principal has a wildcard
        :param principal: 
        :return: 
        """
        if self.debug:
            print('wildcard'+lineno())

            print('principal: '+str(principal)+lineno())
            print('type: '+str(type(principal))+lineno())

        if sys.version_info[0] < 3:
            if type(principal) == type(unicode()):
                return self.has_asterisk(principal)
            elif type(principal) == type(dict()):

                for value in principal:
                    if self.debug:
                        print('value type: ' + str(type(value)) + lineno())
                        print('key: ' + str(value) + lineno())
                        print('value: ' + str(principal[value]) + lineno())

                    if type(value) == type(str()) or type(value) == type(unicode()):
                        if self.has_asterisk(principal[value]):
                            return True

                    elif type(value) == type(list()):
                        for item in value:

                            if self.has_asterisk(item):
                                return True
            return False
        else:
            if type(principal) == type(str()):
                return self.has_asterisk(principal)

            elif type(principal) == type(dict()):

                for value in principal:
                    if self.debug:
                        print('value type: '+str(type(value))+lineno())
                        print('key: '+str(value)+lineno())
                        print('value: '+str(principal[value])+lineno())

                    if type(value) == type(str()) or type(value) == type(unicode()):
                        if self.has_asterisk(principal[value]):
                            return True

                    elif type(value)== type(list()):
                        for item in value:

                            if self.has_asterisk(item):
                                return True
            return False

    def has_asterisk(self, string):
        """
        Whether string has asterisk
        :param string: 
        :return: 
        """
        if self.debug:
            print('has asterisk'+lineno())

        if '*' in string:
            if self.debug:
                print('has an asterisk '+lineno())

            return True

        return False


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - S3BucketPublicReadWriteAclRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class S3BucketPublicReadWriteAclRule(BaseRule):
  
  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)
      
  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
      print('rule_text'+lineno())
    return 'S3 Bucket should not have a public read-write acl'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::FAILING_VIOLATION'
    return 'VIOLATION::FAILING_VIOLATION'

  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
      print('rule_id'+lineno())
    self.id ='F14'
    return 'F14'


  def audit_impl(self):
    """
    Audit
    :return: violations 
    """
    if self.debug:
      print('S3BucketPublicReadWriteAclRule - audit_impl'+lineno())
    logical_resource_ids = []

    resources = self.cfn_model.resources_by_type('AWS::S3::Bucket')

    if len(resources) > 0:
      for resource in resources:
        if self.debug:
          print('resource: ' + str(resource)+lineno())
          print('vars: '+str(vars(resource))+lineno())
        if hasattr(resource,'accessControl'):
          if resource.accessControl:
            if resource.accessControl == 'PublicReadWrite':
              logical_resource_ids.append(str(resource.logical_resource_id))

    else:
      if self.debug:
        print('no violating_policies' + lineno())

    if self.debug:
      input('Press enter to continue: '+lineno())

    return logical_resource_ids


import os
import sys
import unittest
from collections import OrderedDict
from terraform_model.model.EC2SecurityGroupEgress import EC2SecurityGroupEgress as class_to_test


class TestEc2SecurityGroupEgress(unittest.TestCase):
    """
    Test ipaddr
    """

    def test_new_interface(self):


      cfn_model = {

          "Resources": {
              "myENI": {
                  "Type": "AWS::EC2::NetworkInterface",
                  "Properties": {
                      "Tags": [{"Key": "foo", "Value": "bar"}],
                      "Description": "A nice description.",
                      "SourceDestCheck": "false",
                      "GroupSet": ["sg-75zzz219"],
                      "SubnetId": "subnet-3z648z53",
                      "PrivateIpAddress": "10.0.0.16"
                  }
              }
          }
        }

      expected_result = 'EC2SecurityGroupEgress'

      real_result = class_to_test(cfn_model=cfn_model)

      print('expected results: '+str(expected_result))
      print('real results: '+str(real_result))
      print('type: '+str(type(real_result)))
      print('name: '+str(real_result.__class__.__name__))
      self.maxDiff = None
      self.assertEqual(expected_result, real_result.__class__.__name__)




from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)


def lineno():
    """Returns the current line number in our program."""
    return str(' -  Error- line number: '+str(inspect.currentframe().f_back.f_lineno))


class Error(Exception):
   """
   Base class for other exceptions
   """

   def __init__(self, message, debug=False):
       """
       Initialize
       :param message: 
       :param debug: 
       """
       self.debug = debug

       if self.debug:
           print('__init__ - Error'+lineno())

       try:
           ln = sys.exc_info()[-1].tb_lineno
       except AttributeError:
           ln = inspect.currentframe().f_back.f_lineno
       self.args = "{0.__name__} (line {1}): {2}".format(type(self), ln, message),
       
       # FIXME
       sys.exit(self)



from __future__ import absolute_import, division, print_function
import copy
import inspect
from builtins import (str)
from terraform_model.model.References import References
from terraform_model.model.Parameter import Parameter
from terraform_model.model.Data import Data
from terraform_model.model.Output import Output
from terraform_model.model.Locals import Locals
from terraform_model.model.Provider import Provider




def lineno():
    """Returns the current line number in our program."""
    return str(' - CfnModel - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


class CfnModel:
    """
    Cloudformation model
    """

    def __init__(self, debug=False):
        """
        Initialize
        :param debug: 
        """
        # attr_accessor :raw_model
        self.outputs = {}
        self.locals = {}
        self.providers = {}
        self.parameters={}
        self.data = {}
        self.resources = {}
        self.raw_model = None
        self.debug = debug
        if self.debug:
            print('CfnModel - init'+lineno())

    def copy(self):
        """
        copy the model
        :return: copy of model
        """
        if self.debug:
            print('CfnModel - copy'+lineno())

        return copy.copy(self.raw_model)

    def security_groups(self):
        """
        Get security groups
        :return: 
        """
        if self.debug:
            print("\n\n################################################################")
            print('CfnModel - security_groups - getting security group resources'+lineno())
            print("####################################################################\n")

        return self.resources_by_type('AWS::EC2::SecurityGroup')

    def iam_users(self):
        """
        Get iam users
        :return: 
        """
        if self.debug:
            print("\n\n################################################################")
            print('CfnModel - iam_users - getting iam users resources'+lineno())
            print("####################################################################\n")

        return self.resources_by_type('AWS::IAM::User')

    def standalone_ingress(self):
        """
        Get standalone ingress resources
        :return: 
        """
        if self.debug:
            print("\n\n################################################################")
            print('CfnModel - standalone_ingress - getting security group ingress resources'+lineno())
            print("####################################################################\n")

        security_group_ingresses = []

        resources = self.resources_by_type('AWS::EC2::SecurityGroupIngress')

        for resource in resources:
            if self.debug:
                print("\n\n#############################################")
                print('Stand alone ingress'+lineno())
                print(str(resource) + lineno())
                print("################################################\n")

            if 'Properties' in resource.cfn_model:
                if self.debug:
                    print('properties in cfn_model: '+lineno())
                if 'GroupId' in resource.cfn_model['Properties']:
                    if self.debug:
                        print('groupid in properties: '+lineno())

                    if References.is_security_group_id_external(str(resource.cfn_model['Properties']['GroupId']) ,debug=self.debug):
                        security_group_ingresses.append(resource)
        if self.debug:
            print("\n############################################")
            print('These are the standalone security_group_ingresses: '+str(security_group_ingresses)+lineno())
            print("##############################################\n")

        return security_group_ingresses

    def standalone_egress(self):
        """
        Get standalone egress resources
        :return: 
        """
        if self.debug:
            print("\n\n################################################################")
            print('CfnModel - standalone_egress - getting security group egress resources'+lineno())
            print("####################################################################\n")

        security_group_egresses = []

        resources = self.resources_by_type('AWS::EC2::SecurityGroupEgress')

        for resource in resources:
            if self.debug:
                print(str(resource) + lineno())

            if 'variable' in resource.cfn_model:

                if self.debug:
                    print('variables in cfn_model '+lineno())

                if 'GroupId' in resource.cfn_model['variable']:
                    if self.debug:
                        print('GroupId in properties'+lineno())

                    if References.is_security_group_id_external(
                            resource.cfn_model['Properties']['GroupId'],
                            debug=self.debug):

                        security_group_egresses.append(resource)

                if 'groupId' in resource.cfn_model['Properties']:
                    if self.debug:
                        print('groupId in properties'+lineno())

                    if References.is_security_group_id_external(
                            resource.cfn_model['Properties']['groupId'],
                            debug=self.debug):

                        security_group_egresses.append(resource)

        if self.debug:
            print('security_group_egresses: '+str(security_group_egresses)+lineno())

        return security_group_egresses

    def resources_by_type(self, resource_type):
        """
        Get cfn resources by type
        :param resource_type: 
        :return: 
        """
        if self.debug:

            print('CfnModel - resource_by_type'+lineno())
            print("\n\n####################################")
            print('#### Looking for resource_type: '+str(resource_type)+' in raw_model'+lineno())
            print('#### raw model: '+str(self.raw_model)+lineno())
            print("####################################\n\n")

        original_resource_type = resource_type

        resource_map = {
            'AWS::ElasticLoadBalancing::LoadBalancer':'aws_elb',
            'AWS::EC2::SecurityGroup':'aws_security_group',
            'AWS::EC2::Instance':'aws_instance',
            'AWS::IAM::Role': 'aws_iam_role',
            "AWS::IAM::Policy": 'aws_iam_policy',
            "AWS::IAM::User": 'aws_iam_user',
            "AWS::EC2::NetworkInterface": 'aws_network_interface',
            "AWS::IAM::Group": 'aws_iam_group',
            "AWS::S3::BucketPolicy": 'aws_s3_bucket_policy',
            "AWS::SQS::QueuePolicy":'aws_sqs_queue_policy',
            "AWS::IAM::ManagedPolicy":'aws_iam_policy_attachment'
        }

        if resource_type in resource_map:
            resource_type = resource_map[resource_type]

        if self.debug:
            print('trying to find: '+str(resource_type)+lineno())
            input('enter'+lineno())

        resources = []

        if self.debug:
            print(str(self.resources)+lineno())

        # Iterating through the resources in the raw_model
        for resource in self.resources:
            if self.debug:
                print("\n################# Resource Details ########################")
                print('resource: '+str(resource)+lineno())
                print('resource object: '+str(self.resources[resource])+lineno())

                print('type: '+str(self.resources[resource].resource_type)+lineno())
                print('vars: '+str(vars(self.resources[resource]))+lineno())
                print('resource type is: '+str(self.resources[resource].resource_type)+lineno())
                print("############################################################\n")

            if str(self.resources[resource].resource_type) == str(resource_type):
                if self.debug:
                    print(' ### FOUND MATCHING RESOURCE TYPE '+lineno())

                resources.append(self.resources[resource])

        if self.debug:
            print('CfnModel - resources_by_type - returning resource'+lineno())

        if len(resources)<1:
            if self.debug:
                print('### Could not find matching type for: '+str(resource_type)+lineno())
        else:
            if self.debug:
                print("\n\n########################################")
                print('### found '+str(len(resources))+' '+str(resource_type)+' resources'+lineno())
                print("########################################\n\n")

        return resources

    def find_security_group_by_group_id(self, security_group_reference):
        """
        Get security group by security group id
        :param security_group_reference: 
        :return: 
        """
        if self.debug:
            print('CfnModel - find_security_group_by_group_id'+lineno())
            print('security_group_reference: '+str(security_group_reference)+lineno())

        security_group_id = References.resolve_security_group_id(security_group_reference)

        if not security_group_id:
            # # leave it alone since external ref or something we don't grok
            return security_group_reference
        else:

            security_groups = self.security_groups()

            for sg in security_groups:
                if self.debug:
                    print('sg: '+str(sg)+lineno())
                    print('vars: '+str(vars(sg))+lineno())

                if sg.logical_resource_id == sg:
                    return sg

            # leave it alone since external ref or something we don't grok
            return security_group_reference

    def transform_hash_into_parameters(self, cfn_hash):
        """
        Transform hash into parameters
        :param cfn_hash:
        :return:
        """

        if self.debug:
            print('CfnParser - transform_hash_into_parameters'+lineno())
            print('cfn_hash: '+str(cfn_hash)+lineno())


        if 'variable' in cfn_hash and cfn_hash['variable']:

            for param in cfn_hash['variable']:

                if self.debug:
                    print(param+lineno())
                    print(str(cfn_hash['Parameters'][param])+lineno())

                parameter = Parameter(debug=self.debug)
                parameter.id = param
                parameter.type = cfn_hash['variable'][param]['Type']

                parameter.instance_variables.append(param+'='+cfn_hash['variable'][param]['Type'].lower().replace('-','_'))

                self.parameters[param] = parameter

        if 'data' in cfn_hash and cfn_hash['data']:

            for param in cfn_hash['data']:

                if self.debug:
                    print(param+lineno())
                    print(str(cfn_hash['data'][param])+lineno())

                data = Data(debug=self.debug)
                data.id = param
                data.type = cfn_hash['data'][param]['Type']

                data.instance_variables.append(param+'='+cfn_hash['data'][param]['Type'].lower().replace('-','_'))

                self.data[param] = data

        if 'output' in cfn_hash and cfn_hash['output']:

            for param in cfn_hash['output']:

                if self.debug:
                    print(param+lineno())
                    print(str(cfn_hash['output'][param])+lineno())

                output = Output(debug=self.debug)
                output.id = param
                output.type = cfn_hash['output'][param]['Type']

                output.instance_variables.append(param+'='+cfn_hash['output'][param]['Type'].lower().replace('-','_'))

                self.outputs[param] = output

        if 'locals' in cfn_hash and cfn_hash['locals']:

            for param in cfn_hash['locals']:

                if self.debug:
                    print(param+lineno())
                    print(str(cfn_hash['output'][param])+lineno())

                locals = Locals(debug=self.debug)
                locals.id = param
                locals.type = cfn_hash['locals'][param]['Type']

                locals.instance_variables.append(param+'='+cfn_hash['locals'][param]['Type'].lower().replace('-','_'))

                self.locals[param] = locals

        if 'provider' in cfn_hash and cfn_hash['provider']:

            for param in cfn_hash['provider']:

                if self.debug:
                    print(param+lineno())
                    print(str(cfn_hash['output'][param])+lineno())

                locals = Locals(debug=self.debug)
                locals.id = param
                locals.type = cfn_hash['provider'][param]['Type']

                locals.instance_variables.append(param+'='+cfn_hash['provider'][param]['Type'].lower().replace('-','_'))

                self.locals[param] = locals


from __future__ import absolute_import, division, print_function
import inspect
import sys
import json
from terraform_validator.RuleDefinition import RuleDefinition
from collections import OrderedDict
from builtins import (str)



def lineno():
    """Returns the current line number in our program."""
    return str(' - Violation - line number: '+str(inspect.currentframe().f_back.f_lineno))



class Violation(RuleDefinition):

    def __init__(self, id, type, message, logical_resource_ids=None, debug=False):
        """
        Initialize Violation
        :param id: 
        :param type: 
        :param message: 
        :param logical_resource_ids: 
        :param debug: 
        """
        RuleDefinition.__init__(self, id, type, message)
        self.attr_reader = None
        self.message = message
        self.type = type
        self.logical_resource_ids= logical_resource_ids
        self.id=id
        self.debug = debug

        if self.debug:
            print('Violation - init'+lineno())


    def id(self):
        return self.id()

    def type(self):
        return self.type()

    def message(self):
        return self.message()

    def logical_resource_ids(self):
        return self.logical_resource_ids()

    def to_string(self):
        """
        Returns violation as a string
        :return: 
        """
        if self.debug:
            print('to string'+lineno())
        #FIXME
        sys.exit(1)
        #"#{super} #{@logical_resource_ids}"


    def to_hash(self):
        """
        Converts violation to hash
        :return: 
        """
        if self.debug:
            print('to hash'+lineno())
            print('logical resource ids: '+str(self.logical_resource_ids)+lineno())
            print('logical id type: '+str(type(self.logical_resource_ids))+lineno())
            print('message: '+str(self.message)+lineno())




        hash = {'id': self.id,'type':self.type ,'message': self.message, 'logical_resource_ids': str(self.logical_resource_ids)}

        order_of_keys = ["id", "type", "message","logical_resource_ids"]

        list_of_tuples = []
        for key in order_of_keys:
            if self.debug:
                print('key: '+str(key)+lineno())
                print('value: '+str(hash[key])+lineno())

            if str(key) == 'logical_resource_ids' and type(hash[key])== type(list()):
                if self.debug:
                    print('is a list: '+lineno())
                mylist = sorted(json.loads(str(hash[key]).replace("'",'"')))
                my_newlist = []
                for item in mylist:
                    my_newlist.append(str(item))

                list_of_tuples.append(tuple((key, str(my_newlist))))


            else:
                list_of_tuples.append(tuple((key, hash[key])))

        #list_of_tuples = [(key, hash[key]) for key in order_of_keys]
        new_results = OrderedDict(list_of_tuples)



        if self.debug:
            print('#####################################')
            print('hash is: '+str(new_results)+lineno())
            print('#####################################')

        return new_results




from __future__ import absolute_import, division, print_function
import sys

class Output:
    """
    Parameter model
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug:
        """
        #   attr_accessor :id, :type
        # attr_accessor :synthesized_value
        self.id = None
        self.type = None
        self.instance_outputs = []
        self.debug= debug

        if self.debug:
            print('Parameter - init')

    def is_no_echo(self):
        """
        ???
        :return:
        """
        if self.debug:
            print('is_no_echo')
        # FIXME
        sys.exit(1)
        #!@noEcho.nil? && @noEcho.to_s.downcase == 'true'


    def to_string(self):
        """
        ???
        :return:
        """
        if self.debug:
            print('to_s')
        # FIXME
        sys.exit(1)
        #    << END
        #{
        #    # {emit_instance_vars}
        #}

    def method_missing(self, method_name, *args):
        """
        ???
        :param method_name:
        :param args:
        :return:
        """
        if self.debug:
            print('method_missing')
        # FIXME
        sys.exit(1)
        #if method_name =~ /^(\w+)=$/
        #  instance_variable_set "@#{$1}", args[0]
        #else
        #  instance_variable_get "@#{method_name}"


    def emit_instance_vars(self):
        """
        ???
        :return:
        """
        if self.debug:
            print('emit_instance_varts')

        # FIXME
        sys.exit(1)
        #instance_vars_str = ''
        #self.instance_variables.each do |instance_variable|
        #  instance_vars_str += "  #{instance_variable}=#{instance_variable_get(instance_variable)}\n"
        #end
        #instance_vars_str


"""
The command line interface to terraform_validator.

"""
from __future__ import absolute_import, division, print_function
import sys
import click
import inspect
import subprocess
import terraform_validator
from terraform_validator import ValidateUtility
from terraform_validator.RuleDumper import RuleDumper


def lineno():
    """Returns the current line number in our program."""
    return str(' - ValidateUtility - line number: '+str(inspect.currentframe().f_back.f_lineno))


@click.group()
@click.version_option(version='0.6.40')
def cli():
    pass


@cli.command()
@click.option('--pypi-update-check/--no-pypi-update-check', help='Whether to enable pypi check', default=False, required=False)
@click.option('--s3-profile', is_flag=False, default='default', help="AWS profile")
@click.option('--s3-bucket-name', is_flag=False ,help='S3 Bucket name for custom rules')
@click.option('--excluded-rules', is_flag=False,default='',help='Comma separated string of rules to exclude')
@click.option('--suppress-errors','-s',help='Suppress warnings like bad file format, etc', required=False, is_flag=True)
@click.option('--template-path', '-t', help='base directory to search for templates', required=False)
@click.option('--debug',help='Turn on debugging', required=False, is_flag=True)
@click.option('--rules-directory', '-r', help='Extra rule directory', required=False)
@click.option('--profile-path', '-o', help='Path to a profile file', required=False)
@click.option('--allow-suppression/--no-allow-suppression', help='Allow using Metadata to suppress violations', default=True, required=False)
@click.option('--print-suppression','-p', help='Emit suppressions to stderr', required=False, is_flag=True)
@click.option('--parameter-values-path', '-m', help='Path to a JSON file to pull Parameter values from', required=False)
@click.option('--isolate-custom-rule-exceptions', '-i', help='Isolate custom rule exceptions - just emit the exception without stack trace and keep chugging',is_flag=True)
@click.option('--version', '-v', help='Print version and exit', required=False, is_flag=True)
@click.option('--use-optional-rules',help='Use optional rules', required=False, is_flag=True)
def validate(suppress_errors,
             template_path,
             debug,rules_directory,
             profile_path,
             allow_suppression,
             print_suppression,
             parameter_values_path,
             isolate_custom_rule_exceptions,
             version,
             use_optional_rules,
             excluded_rules,
             s3_bucket_name,
             s3_profile,
             pypi_update_check
             ):
    '''
    primary function for validating a template
    :param template_path:
    :param template_file:
    :param debug:
    :param rules_directory:
    :param profile_path:
    :param allow_suppression:
    :param print_suppression:
    :param parameter_values_path:
    :param isolate_custom_rule_exceptions:
    :param version:
    :param use_optional_rules
    :param excluded_rules
    :param s3_bucket_name
    :param s3_profile
    :param pypi_update_check
    :return:
    '''

    if debug:
        debug = True
    else:
        debug = False

    if version:
        myversion()
    else:

        if not template_path and not template_file:
            print('Must have either an input_path or template_file')
            sys.exit(1)

        excluded_rules = excluded_rules.split(",")

        new_excluded_rules = []
        for items in excluded_rules:
            j = items.replace(' ', '')
            new_excluded_rules.append(j)

        start_validate(
            suppress_errors,
            template_path,
            debug,
            rules_directory,
            profile_path,
            allow_suppression,
            print_suppression,
            parameter_values_path,
            isolate_custom_rule_exceptions,
            use_optional_rules,
            new_excluded_rules,
            s3_bucket_name,
            s3_profile,
            pypi_update_check
        )


@cli.command()
@click.option('--profile-path', '-o', help='Path to a profile file', required=False)
@click.option('--rule-directory', '-r', help='Extra rule directory', required=False)
@click.option('--debug',help='Turn on debugging', required=False, is_flag=True)
def dump_rules(debug, rule_directory, profile_path):
    """
    dump rules
    :param debug:
    :param rule_directory:
    :param profile_path:
    :return:
    """
    start_dump_rules(rule_directory=rule_directory, debug=debug, profile_definition=profile_path)

@click.option('--version', '-v', help='Print version and exit', required=False, is_flag=True)
def version(version):
    """
    Get version
    :param version:
    :return:
    """
    myversion()


def start_dump_rules(rule_directory, profile_definition, debug):
    """
    start dumping rules
    :param rule_directory:
    :param profile_definition:
    :param debug:
    :return:
    """
    dumper = RuleDumper(profile_definition=profile_definition, rule_directory=rule_directory, debug=debug)
    dumper.dump_rules()



def myversion():
    '''
    Gets the current version
    :return: current version
    '''
    print('Version: ' + str(terraform_validator.__version__))

def start_validate(
        suppress_errors,
        template_path,
        debug,
        rules_directory,
        profile_path,
        allow_suppression,
        print_suppression,
        parameter_values_path,
        isolate_custom_rule_exceptions,
        use_optional_rules,
        excluded_rules,
        s3_bucket_name,
        s3_profile,
        pypi_update_check):
    '''
    Starts the validation
    :param template_path:
    :param template_file:
    :param debug:
    :param rules_directory:
    :param profile_path:
    :param allow_suppression:
    :param print_suppression:
    :param parameter_values_path:
    :param isolate_custom_rule_exceptions:
    :param use_optional_rules
    :param s3_bucket_name
    :param s3_profile
    :return:
    '''
    if debug:
        print('command - start_validate'+lineno())
        print('input_path: '+str(template_path))
        print('pypi update checK: '+str(pypi_update_check))

    if pypi_update_check:
        check_for_updates(debug=debug)


    config_dict = {}
    config_dict['suppress_errors']=suppress_errors
    config_dict['input_path'] = template_path
    config_dict['debug'] = debug
    config_dict['rules_directory'] = rules_directory
    config_dict['profile'] = profile_path
    config_dict['allow_suppression'] = allow_suppression
    config_dict['print_suppression'] = print_suppression
    config_dict['parameter_values_path'] = parameter_values_path
    config_dict['isolate_custom_rule_exceptions'] = isolate_custom_rule_exceptions
    config_dict['use_optional_rules'] = use_optional_rules
    config_dict['excluded_rules'] = excluded_rules
    config_dict['s3_bucket_name'] = s3_bucket_name
    config_dict['s3_profile'] = s3_profile
    validator = ValidateUtility(config_dict)
    if debug:
        print('print have validator')
    if validator.validate():
        if debug:
            print('validated')
    else:
        if debug:
            print('not validated')


def get_current_pip_version(debug=False):

    try:
        command = 'pip search terraform-validator | grep terraform-validator | cut -d\' \' -f 2 |  sed -e "s/^(//" -e "s/)//"'
        if debug:
            print('command: '+str(command))
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = process.communicate()
        if stdout:
            current_pip_version = str(stdout).rstrip('\n').strip()
            return current_pip_version
        if stderr:
            print('Error setting up build number')
            elines = stderr.splitlines()
            for e in elines:
                print((e))

    except Exception as e:
        print('Error trying to determine the current terraform-validator version')
        return 0

def check_for_updates(debug=False):

    try:

        if sys.version_info[0] < 3:
            current_pip_version = get_current_pip_version(debug=debug).decode("utf-8")
            current_local_version = str(terraform_validator.__version__).rstrip('\n').strip().decode("utf-8")

        else:
            current_pip_version = get_current_pip_version(debug=debug)
            current_local_version = str(terraform_validator.__version__).rstrip('\n').strip()

        if debug:
            print('current pypi version: '+str(current_pip_version))
            print('current local version: '+str(current_local_version))

        if current_pip_version == 0:
            raise Exception
        elif current_pip_version != current_local_version:
            print('#########################################################################################')
            print('There is a more current version of terraform-validator. You should update ')
            print('terraform-validator with pip install -U terraform-validator')
            print('#########################################################################################')
            sys.exit(1)
        else:
            if debug:
                print('terraform-validator is the most current version')
    except Exception as e:
        print('Error trying to determine the current terraform-validator version')



from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule



def lineno():
    """Returns the current line number in our program."""
    return str(' -  EbsVolumeHasSseRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class EbsVolumeHasSseRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize EbsVolumeHasSseRule
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Returns rule text
        :return:
        """
        if self.debug:
          print('rule_text'+lineno())
        return 'EBS volume should have server-side encryption enabled'

    def rule_type(self):
        """
        Returns rule type
        :return:
        """
        self.type= 'VIOLATION::FAILING_VIOLATION'
        return 'VIOLATION::FAILING_VIOLATION'

    def rule_id(self):
        """
        Returns rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='F1'
        return 'F1'


    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('EbsVolumeHasSseRule - audit_impl'+lineno())

        violating_volumes = []

        resources = self.cfn_model.resources_by_type('AWS::EC2::Volume')

        if len(resources) > 0:

            for resource in resources:
                if self.debug:
                    print('resource: ' + str(resource)+lineno())
                    print('vars: '+str(vars(resource)))

                if hasattr(resource, 'encrypted'):
                    if self.debug:
                        print('has encrypted attribute'+lineno())
                    if resource.encrypted == 'true' or resource.encrypted == 'True':
                        if self.debug:
                            print('it is true')
                    else:
                        if self.debug:
                            print('not true')
                        violating_volumes.append(str(resource.logical_resource_id))
                else:
                    if self.debug:
                        print('does not have encrypted property')
                    violating_volumes.append(str(resource.logical_resource_id))

        else:
            if self.debug:
                print('no violating_volumes'+lineno())

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_volumes


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' - UserHasInlinePolicyRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))

class UserHasInlinePolicyRule(BaseRule):


  def __init__(self, cfn_model=None, debug=None):
      """
      Initialize
      :param cfn_model:
      """
      BaseRule.__init__(self, cfn_model, debug=debug)
      
  def rule_text(self):
      """
      Get rule text
      :return:
      """
      if self.debug:
          print('rule_text'+lineno())
      return 'IAM user should not have any inline policies.  Should be centralized Policy object on group'


  def rule_type(self):
      """
      Get rule type
      :return:
      """
      self.type= 'VIOLATION::FAILING_VIOLATION'
      return 'VIOLATION::FAILING_VIOLATION'


  def rule_id(self):
      """
      Get rule id
      :return:
      """
      if self.debug:
          print('rule_id'+lineno())
      self.id ='F10'
      return 'F10'


  def audit_impl(self):
      """
      Audit
      :return: violations
      """
      if self.debug:
          print('UserHasInlinePolicyRule - audit_impl'+lineno())

      violating_users = []

      for user in self.cfn_model.iam_users():
          if self.debug:
              print('user: '+str(user))
              print('vars: '+str(vars(user))+lineno())
          if hasattr(user,'policy_objects'):
              if user.policy_objects:
                  if len(user.policy_objects)>0:
                      violating_users.append(str(user.logical_resource_id))

      if self.debug:
          input('Press enter to continue: ' + lineno())

      return violating_users


from __future__ import absolute_import, division, print_function
from terraform_model.model.ModelElement import ModelElement


class EC2NetworkInterface(ModelElement):
    """
    Ec2 network interface model lement
    """
    def __init__(self, cfn_model):
        """
        Initialize
        :param cfn_model: 
        """
        ModelElement.__init__(self, cfn_model)

        self.groupSet= []
        self.ipv6Addresses= []
        self.privateIpAddresses= []
        self.tags= []
        self.security_groups= []
        self.resource_type = 'AWS::EC2::NetworkInterface'






from __future__ import absolute_import, division, print_function
import inspect
from terraform_model.parser.SecurityGroupParser import SecurityGroupParser
from terraform_model.parser.Ec2NetworkInterfaceParser import Ec2NetworkInterfaceParser
from terraform_model.parser.Ec2InstanceParser import Ec2InstanceParser
from terraform_model.parser.LoadBalancerParser import LoadBalancerParser
from terraform_model.parser.LoadBalancerV2Parser import LoadBalancerV2Parser
from terraform_model.parser.IamGroupParser import IamGroupParser
from terraform_model.parser.IamUserParser import IamUserParser
from terraform_model.parser.IamRoleParser import IamRoleParser
from terraform_model.parser.SecurityGroupParser import SecurityGroupParser
from terraform_model.parser.WithPolicyDocumentParser import WithPolicyDocumentParser

def lineno():
    """Returns the current line number in our program."""
    return str(' -  ParserRegistry- line number: '+str(inspect.currentframe().f_back.f_lineno))


class ParserRegistry(object):
    """
    Parser registry
    """
    def __init__(self, debug=False):
        """
        Initialize
        :param debug: 
        """
        self.debug = debug

        if self.debug:
            print('ParserRegistry - init'+lineno())

        self.registry= {
            'AWS::EC2::SecurityGroup' : SecurityGroupParser,
            'AWS::EC2::NetworkInterface' : Ec2NetworkInterfaceParser,
            'AWS::EC2::Instance' : Ec2InstanceParser,
            'AWS::ElasticLoadBalancing::LoadBalancer' : LoadBalancerParser,
            'AWS::ElasticLoadBalancingV2::LoadBalancer' : LoadBalancerV2Parser,
            'AWS::IAM::Group' : IamGroupParser,
            'AWS::IAM::User' : IamUserParser,
            'AWS::IAM::Role' : IamRoleParser,
            'AWS::IAM::Policy' : WithPolicyDocumentParser,
            'AWS::IAM::ManagedPolicy' : WithPolicyDocumentParser,
            'AWS::S3::BucketPolicy' : WithPolicyDocumentParser,
            'AWS::SNS::TopicPolicy' : WithPolicyDocumentParser,
            'AWS::SQS::QueuePolicy' : WithPolicyDocumentParser
        }


from __future__ import absolute_import, division, print_function
import sys
import copy
import inspect
from terraform_model.model.Policy import Policy
from terraform_model.parser.PolicyDocumentParser import PolicyDocumentParser

def lineno():
    """Returns the current line number in our program."""
    return str(' -  IamGroupParser- line number: '+str(inspect.currentframe().f_back.f_lineno))

class IamGroupParser:
    """
    IAM group parser
    """
    
    @staticmethod
    def parse(cfn_model, resource, debug=False):
        """
        Parse iam group
        :param resource: 
        :param debug: 
        :return: 
        """
        if debug:
            print('IAMGroupParser - parse'+lineno())
            print('resource: '+str(resource)+lineno())

        iam_group = copy.copy(resource)

        if debug:
            print('attributes: '+str(vars(iam_group))+lineno())

        for policy in iam_group.policies:
            if debug:
                print('policy: '+str(policy)+lineno())

            if 'Policies' in policy:
                if len(policy['Policies'])>0:

                    new_policy = Policy(debug=debug)
                    new_policy.name = policy['PolicyName']
                    new_policy.policy_document = PolicyDocumentParser.parse(policy['PolicyDocument'])

                    iam_group.policy_objects.append(new_policy)

        return iam_group


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' - SecurityGroupEgressPortRangeRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class SecurityGroupEgressPortRangeRule(BaseRule):


  def __init__(self, cfn_model=None, debug=None):
    """
    Initialize
    :param cfn_model: 
    """
    BaseRule.__init__(self, cfn_model, debug=debug)

  def rule_text(self):
    """
    Get rule text
    :return: 
    """
    if self.debug:
        print('rule_text'+lineno())
    return 'Security Groups found egress with port range instead of just a single port'


  def rule_type(self):
    """
    Get rule type
    :return: 
    """
    self.type= 'VIOLATION::WARNING'
    return 'VIOLATION::WARNING'


  def rule_id(self):
    """
    Get rule id
    :return: 
    """
    if self.debug:
        print('rule_id'+lineno())
    self.id ='W29'
    return 'W29'


  ##
  # This will behave slightly different than the legacy jq based rule which was targeted against inline ingress only
  def audit_impl(self):
    """
    Audit
    :return: violations 
    """
    if self.debug:
        print('SecurityGroupEgressPortRangeRule - audit_impl'+lineno())
    violating_egresses = []

    for groups in self.cfn_model.security_groups():

        if self.debug:
            print('group: '+str(groups)+lineno())

        if type(groups)==type(dict()):
            if str(groups['FromPort']) != str(groups['ToPort']):
                violating_egresses.append(groups.logical_resource_id)

        else:

            for egress in groups.egresses:
              if self.debug:
                print('## egresses: '+str(egress))

              if type(egress)==type(dict()):

                if str(egress['FromPort']) != str(egress['ToPort']):
                    violating_egresses.append(str(groups.logical_resource_id))

              elif type(egress) == type(list()):

                  for item in egress:

                      if type(item) == type(dict()):

                          if str(item['FromPort']) != str(item['ToPort']):
                              violating_egresses.append(groups.logical_resource_id)

                      elif hasattr(item,'fromPort') and hasattr(item,'toPort') and item.fromPort != item.toPort:
                          violating_egresses.append(str(item.logical_resource_id))

              elif hasattr(egress,'fromPort') and hasattr(egress,'toPort') and egress.fromPort != egress.toPort:
                violating_egresses.append(str(egress.logical_resource_id))

    violating_standalone_egresses = self.cfn_model.standalone_egress()
    if self.debug:
        print('violating standalone egresses: '+str(violating_standalone_egresses)+lineno())

    # For standalone egress resource types
    for groups in violating_standalone_egresses:
        if self.debug:
            print('group: '+str(groups)+lineno())
            print('vars: '+str(vars(groups))+lineno())

        if hasattr(groups,'cfn_model'):
            if 'Properties' in groups.cfn_model:
                if 'ToPort' in groups.cfn_model['Properties'] and 'FromPort' in groups.cfn_model['Properties']:

                    if str(groups.cfn_model['Properties']['ToPort']) != str(groups.cfn_model['Properties']['FromPort']):
                        violating_egresses.append(str(groups.logical_resource_id))

    if self.debug:
      input('Press enter to continue: '+lineno())

    return violating_egresses


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule

def lineno():
    """Returns the current line number in our program."""
    return str(' - SqsQueuePolicyNotActionRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class SqsQueuePolicyNotActionRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Get rule text
        :return:
        """
        if self.debug:
          print('rule_text'+lineno())
        return 'SQS Queue policy should not allow Allow+NotAction'


    def rule_type(self):
        """
        Get rule type
        :return:
        """
        self.type= 'VIOLATION::WARNING'
        return 'VIOLATION::WARNING'


    def rule_id(self):
        """
        Get rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='W18'
        return 'W18'

    def audit_impl(self):
      """
      Audit
      :return: violations
      """
      if self.debug:
        print('SqsQueuePolicyNotActionRule - audit_impl'+lineno())

      violating_policies = []

      resources = self.cfn_model.resources_by_type('AWS::SQS::QueuePolicy')

      if len(resources)>0:
        for resource in resources:
            if self.debug:
              print('resource: '+str(resource)+lineno())

            if hasattr(resource, 'policy_document'):

              if resource.policy_document:
                if self.debug:
                  print('has policy document ' + lineno())
                  print(resource.policy_document.statements)
                if resource.policy_document.allows_not_action():
                  if self.debug:
                    print('has allows not action?')

                  violating_policies.append(str(resource.logical_resource_id))
      else:
        if self.debug:
          print('no violating_policies' + lineno())

      if self.debug:
          input('Press enter to continue: ' + lineno())

      return violating_policies


from __future__ import absolute_import, division, print_function
import copy
import sys
import inspect
from builtins import (str)
from terraform_validator.custom_rules.BaseRule import BaseRule


def lineno():
    """Returns the current line number in our program."""
    return str(' - SecurityGroupIngressPortRangeRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))



class SecurityGroupIngressPortRangeRule(BaseRule):

    def __init__(self, cfn_model=None, debug=None):
        """
        Initialize
        :param cfn_model:
        """
        BaseRule.__init__(self, cfn_model, debug=debug)

    def rule_text(self):
        """
        Get rule text
        :return:
        """
        if self.debug:
            print('rule_text'+lineno())
        return 'Security Groups found ingress with port range instead of just a single port'


    def rule_type(self):
        """
        Get rule type
        :return:
        """
        self.type= 'VIOLATION::WARNING'
        return 'VIOLATION::WARNING'


    def rule_id(self):
        """
        Get rule id
        :return:
        """
        if self.debug:
            print('rule_id'+lineno())
        self.id ='W27'
        return 'W27'



    def audit_impl(self):
        """
        Audit
        :return: violations
        """
        if self.debug:
            print('SecurityGroupIngressPortRangeRule - audit_impl'+lineno())

        violating_ingresses = []

        for groups in self.cfn_model.security_groups():

            if self.debug:
                print("\n#######################################")
                print('group: '+str(groups)+lineno())
                print('vars: '+str(vars(groups))+lineno())
                if groups.ingresses:
                    print('ingresses: '+str(groups.ingresses)+lineno())
                    for ingres in groups.ingresses:
                        print('vars: '+str(vars(ingres))+lineno())
                print("#########################################\n")


            for ingress in groups.ingresses:
                if ingress.from_port and ingress.to_port:
                    if self.debug:
                        print('has from port and to port'+lineno())

                if str(ingress.from_port) != str(ingress.to_port):
                    violating_ingresses.append(ingress.logical_resource_id)


        if self.debug:
            print('violations: '+str(list(set(violating_ingresses))))

        if self.debug:
            input('Press enter to continue: ' + lineno())

        return violating_ingresses


from __future__ import absolute_import, division, print_function
import inspect
import sys
import json
from collections import OrderedDict



def lineno():
    """Returns the current line number in our program."""
    return str(' - RulesView - line number: '+str(inspect.currentframe().f_back.f_lineno))

class  RulesView:
    """
    Rules view
    """

    @staticmethod
    def emit(rule_registry, profile, debug=None):
        """
        Emits something
        :param profile: 
        :return: 
        """
        if debug:
            print('rule_registry: '+str(rule_registry)+lineno())
            print('profile: '+str(profile)+lineno())


        RulesView.emit_warnings(warnings=rule_registry.warnings(),profile=profile, debug=debug)
        RulesView.emit_failings(failings=rule_registry.failings(),profile=profile, debug=debug)

    @staticmethod
    def emit_warnings(warnings, profile, debug=None):
        """
        Emits warnings
        :param profile: 
        :return: 
        """
        if debug:
            print('emit_warnings '+lineno())
            print('warnings: '+str(warnings)+lineno())


        warnings_data={}
        #d_sorted_by_value = OrderedDict(sorted(d.items(), key=lambda x: x[1]))
        for rule in warnings:
            data = rule.to_hash()


            if debug:
                print('data: '+str(data)+lineno())

            warnings_data[data['id']]=data

        keylist = list(warnings_data.keys())

        keylist.sort()

        print("##################################")
        print("########## WARNINGS ##############")
        print("##################################")

        order_of_keys = ["id", "type", "message"]
        #print(warnings_data)
        if warnings_data:
            for warn_key in warnings_data:
                list_of_tuples = [(key, warnings_data[warn_key][key]) for key in order_of_keys]
                new_results = OrderedDict(list_of_tuples)

                print("%s" % json.dumps(new_results))


        # FIXME FOR profile
        #do | warning |
        #if profile.nil?
        #puts
        #"#{warning.id} #{warning.message}"

        #elsif
        #profile.execute_rule?(warning.id)
        #puts
        #"#{warning.id} #{warning.message}"

    @staticmethod
    def emit_failings(failings, profile, debug=None):
        """
        Emits failings
        :param profile: 
        :return: 
        """

        if debug:
            print('emit_failings ' + lineno())
            print('failings: '+str(failings)+lineno())


        failings_data={}
        for rule in failings:
            data = rule.to_hash()

            if debug:
                print('data: '+str(data)+lineno())

            failings_data[data['id']]=data

        keylist = list(failings_data.keys())
        keylist.sort()

        print("##################################")
        print("########## FAILINGS #############")
        print("##################################")

        order_of_keys = ["id", "type", "message"]
        #print(warnings_data)
        if failings_data:
            for fail_key in failings_data:
                list_of_tuples = [(key, failings_data[fail_key][key]) for key in order_of_keys]
                new_results = OrderedDict(list_of_tuples)

                print("%s" % json.dumps(new_results))

        # FIXME for profile
        # elsif
        # profile.execute_rule?(failing.id)
        # puts
        # "#{failing.id} #{failing.message}"





from __future__ import absolute_import, division, print_function
import sys
import inspect
from terraform_validator.Violation import Violation


def lineno():
    """Returns the current line number in our program."""
    return str(' - BaseRule - caller: '+str(inspect.stack()[1][3])+' - line number: '+str(inspect.currentframe().f_back.f_lineno))


# Base class all Rules should subclass
class BaseRule():
    """
    Base rule
    """

    def __init__(self, cfn_model=None, debug=False):
        """
        Initialize BaseRule
        :param cfn_model:
        :param debug:
        """
        self.cfn_model = cfn_model
        self.type = None
        self.id = None
        self.debug = debug

    
    def audit_impl(self):
        """
        Returns a collection of logical resource ids
        :return:
        """
        raise 'must implement in subclass'
    
    
    def audit(self):
        """
        Returns nil when there are no violations
        Returns a Violation object otherwise
        :return:
        """
        if self.debug:
            print("\n\n##########################################")
            print('BaseRule - audit'+lineno())
            print('calling audit_impl on rule')
            print("##############################################\n")
        
        logical_resource_ids =  self.audit_impl()
        
        if logical_resource_ids:
            if self.debug:
                print('logical ids: '+str(logical_resource_ids)+lineno())
        
            violation = Violation(id=self.rule_id(), type= self.rule_type(), message= self.rule_text(), logical_resource_ids= logical_resource_ids, debug=self.debug)
        
            if self.debug:
                print('violation: '+str(vars(violation))+lineno())
        
            return violation
        
        else:
            if self.debug:
                print('no logical ids: '+lineno())
            return


from __future__ import absolute_import, division, print_function
import inspect
import sys
from builtins import (str)


def lineno():
    """Returns the current line number in our program."""
    return str(' - References - line number: '+str(inspect.currentframe().f_back.f_lineno))



class References:
    """
    References model
    """
    
    @staticmethod
    def resolve_value(cfn_model, value, debug=False):
        """
        ???
        :param value:
        :param debug:
        :return:
        """
        if debug:
            print('resolve_value'+lineno())
            print('value: '+str(value)+lineno())

        if type(value) == type(dict()):

            if 'Ref' in value:
                ref_id = value['Ref']

                if sys.version_info[0] < 3:
                    if type(ref_id) == type(unicode()):
                        if str(ref_id) in cfn_model.parameters:
                            return value
                            # return value if cfn_model.parameters[ref_id].synthesized_value.nil?
                            # return cfn_model.parameters[ref_id].synthesized_value
                        else:
                           return value
                    else:
                      return value

                else:
                    if type(ref_id) == type(str()):
                        if debug:
                            print('ref id is a string '+lineno())
                            print('params: '+str(cfn_model))
                        if str(ref_id) in cfn_model.parameters:
                            if debug:
                                print('ref id in parameters - id is: '+str(ref_id)+lineno())
                                print('returning: '+str(value))
                            return value
                            # return value if cfn_model.parameters[ref_id].synthesized_value.nil?
                            # return cfn_model.parameters[ref_id].synthesized_value
                        else:
                           return value
                    else:
                      return value
            else:
                return value
        else:
          return value

    @staticmethod
    def is_security_group_id_external(group_id, debug=False):
        """
        Is security group id external
        :param debug:
        :return:
        """
        if debug:
            print('\n#####################################################')
            print('Checking if security group id is external')
            print('is_security_group_id_external'+lineno())
            print('group_id: '+str(group_id)+lineno())
            print("######################################################\n")

        if group_id:
            id = References.resolve_security_group_id(group_id,debug=debug)

            if id:
                return id
            else:
                return None

        return False

    @staticmethod
    def resolve_security_group_id(group_id, debug=False):
        """
        Resolve security group id
        :param debug:
        :return:
        """
        if debug:
            print("\n\n##########################################")
            print('resolve_security_group id'+lineno())
            print('group_id: '+str(group_id)+lineno())
            print("##############################################\n")

        if group_id:
            if debug:
                print('there is a group id'+lineno())

            if sys.version_info[0] < 3:

                if type(group_id) == type(unicode()):
                    if debug:
                        print('type is unicode' + lineno())

                return group_id

            elif type(group_id) == type(str()):
                if debug:
                    print('type is string'+lineno())

                return group_id


            elif type(group_id) == type(dict()):
                if debug:
                    print('type is dict'+lineno())
                    print('groupid '+str(group_id)+lineno())

                ## an imported value can only yield a literal to an external sg vs. referencing something local
                if 'Ref' in group_id:
                    if debug:
                        print('returning '+str(group_id['Ref'])+lineno())
                    return group_id['Ref']

                elif 'Fn::GetAtt' in group_id:
                    return References.logical_resource_id_from_get_att(group_id['Fn::GetAtt'])
                else:
                    # anything else will be string manipulation functions
                    # which again leads us back to a string which must be an external security group known out of band
                    # so don't/can't link it up to a security group
                    return None

            else:
                if debug:
                    print('type is not a string'+lineno())
                return None
        else:
            return None

    @staticmethod
    def logical_resource_id_from_get_att(attribute_spec, debug=False):
        """
        ???
        :param debug:
        :return:
        """
        if debug:
            print('logical_resource_id_from_get_att'+lineno())
            print('attribute_spec: '+str(attribute_spec)+lineno())

        if type(attribute_spec) == type(list()):
            if attribute_spec[1] == 'GroupId':
                return attribute_spec[0]
            else:
                # this could be a reference to a nested stack output so treat it as external
                # and presume the ingress is freestanding.
                return None
        elif type(attribute_spec) == type(str()) or type(attribute_spec) == type(unicode()):
            attributes = attribute_spec.split('.')

            if attributes[1] == 'GroupId':
                return attributes[0]
            else:
                # this could be a reference to a nested stack output so treat it as external
                # and presume the ingress is freestanding.
                return None


from __future__ import absolute_import, division, print_function
import inspect



def lineno():
    """Returns the current line number in our program."""
    return str(' - Profile - line number: '+str(inspect.currentframe().f_back.f_lineno))

class Profile:
    """
    Profile
    """
    rules_registry = None

    def __init__(self, debug=False):
        """
        Initialize profile
        :param debug:
        """
        self.debug = debug
        self.rule_ids = []

        if self.debug:
            print('Profile - init'+lineno())


    # Add a Rule to a profile
    def add_rule(self, rule_id):
        """
        Add rule to profile
        :param rule_id:
        :return:
        """
        self.rule_ids.append(rule_id)

    def execute_rule(self, rule_id):
        """
        Whether to execute the rule id
        :param rule_id:
        :return: boolean
        """
        if self.debug:
            print('rule_id: '+str(rule_id)+lineno())
            print('rule id list in profile: '+str(self.rule_ids)+lineno())
        if rule_id in self.rule_ids:
            if self.debug:
                print('found rule id in profile rules id array'+lineno())

            return True
        return False



from __future__ import absolute_import, division, print_function
import inspect
import json
from terraform_model.parser import ParserError

def lineno():
    """Returns the current line number in our program."""
    return str(' -  ResourceTypeValidator- line number: '+str(inspect.currentframe().f_back.f_lineno))


class ResourceTypeValidator:
    """
    Resource Type Validator
    """

    def __init__(self, debug=False):
        """
        Initialize
        :param debug:
        """
        self.debug=debug
        if self.debug:
            print('ResourceTypeValidator - init'+lineno())

    def validate(self, cloudformation_yml):
        """
        validate cloud formation
        :param cloudformation_yml:
        :return:
        """

        if self.debug:
            print("\n\n########################################")
            print('ResourceTypeValidator - validate'+lineno())
            print('Iterating through each of the resources and validating them')
            print('cloudformation template: '+str(cloudformation_yml))
            print("############################################\n\n")

        if type(cloudformation_yml) == type(str()):
            json_acceptable_string = cloudformation_yml.replace("'", "\"")
            cloudformation_yml= json.loads(json_acceptable_string)


        if 'resource' not in cloudformation_yml:
            raise ParserError.ParserError('Illegal cfn - no resources')
        else:
            if self.debug:
                print('cloudformation has resource'+str(cloudformation_yml['resource'])+lineno())
                print('type:'+str(type(cloudformation_yml['resource'])))


        # Don't need a type because the resource is the type
        #for resource in cloudformation_yml['resource']:

        #    if self.debug:
        #        print('resource: '+str(resource))

        #    if 'type' not in cloudformation_yml[resource]:
        #        raise ParserError.ParserError('Illegal cfn - missing type: id: '+str(resource))
        #    else:
        #        if self.debug:
        #            print('resource: '+str(resource)+' has type'+lineno())

        if  'variable' in cloudformation_yml:

            if self.debug:
                print('there is a variable in template: '+lineno())

        if self.debug:
            print('returning '+str(cloudformation_yml)+lineno())

        return cloudformation_yml


import os
import sys
import unittest
from contextlib import contextmanager
if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO
    import io

from mock import patch
from collections import OrderedDict
import terraform_validator
from terraform_validator import command as class_to_test


def pretty(value, htchar='\t', lfchar='\n', indent=0):
    nlch = lfchar + htchar * (indent + 1)
    if type(value) == type(dict()) or type(value) == type(OrderedDict()):

        items = [
            nlch + repr(key) + ': ' + pretty(value[key], htchar, lfchar, indent + 1)
            for key in value
        ]
        return '{%s}' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) == type(list()):
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '[%s]' % (','.join(items) + lfchar + htchar * indent)
    elif type(value) is tuple:
        items = [
            nlch + pretty(item, htchar, lfchar, indent + 1)
            for item in value
        ]
        return '(%s)' % (','.join(items) + lfchar + htchar * indent)
    else:
        return repr(value)

@contextmanager
def captured_output():
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class TestCommand(unittest.TestCase):
    """
    Test command class
    """

    @patch('sys.exit')
    @patch('terraform_validator.command.get_current_pip_version')
    def test_check_for_updates_outdated_version(self,version,exit):
        """
        Tests for when pypi version does not match current version
        :param version:
        :param exit:
        :return:
        """
        version.return_value='0.2.34'
        exit.return_value=None
        expected_results = "#########################################################################################\n" \
                  "There is a more current version of terraform-validator. You should update \n" \
                  "terraform-validator with pip install -U terraform-validator\n" \
                  "#########################################################################################"

        with captured_output() as (out, err):
            class_to_test.check_for_updates()
        # This can go inside or outside the `with` block
        output = out.getvalue().strip()

        real_result = class_to_test.check_for_updates()

        self.maxDiff = None
        self.assertEqual(exit.return_value, real_result)

        self.assertEqual(output,expected_results)




    @patch('terraform_validator.command.get_current_pip_version')
    def test_check_for_updates_current_version(self,version):
        """
        Test for when pypi version matches current version
        :param version:
        :return:
        """
        version.return_value=terraform_validator.__version__

        with captured_output() as (out, err):
            class_to_test.check_for_updates()
        # This can go inside or outside the `with` block
        output = out.getvalue().strip()

        real_result = class_to_test.check_for_updates()

        self.maxDiff = None
        self.assertEqual(output,'')


    @patch('sys.exit')
    @patch('terraform_validator.command.get_current_pip_version')
    def test_check_for_updates_error(self,version,exit):
        """
        Testing for when there is an error retrieving info from pypi
        :param version:
        :param exit:
        :return:
        """
        version.return_value=0
        exit.return_value=None
        expected_results = "Error trying to determine the current terraform-validator version"

        with captured_output() as (out, err):
            class_to_test.check_for_updates()
        # This can go inside or outside the `with` block
        output = out.getvalue().strip()


        real_result = class_to_test.check_for_updates()

        self.maxDiff = None
        self.assertEqual(exit.return_value, real_result)

        self.assertEqual(output,expected_results)




from __future__ import absolute_import, division, print_function
import inspect
import sys
from terraform_model.model.IAMPolicy import IAMPolicy
from terraform_model.model.PolicyDocument import PolicyDocument
from terraform_model.parser.PolicyDocumentParser import PolicyDocumentParser


def lineno():
    """Returns the current line number in our program."""
    return str(' -  WithPolicyDocumentParser- line number: '+str(inspect.currentframe().f_back.f_lineno))

class WithPolicyDocumentParser:
    """
    With policy document parser
    """
    @staticmethod
    def parse(cfn_model, resource, debug=False):
        """
        Parse with policy document parser
        :param resource:
        :param debug:
        :return:
        """
        if debug:
            print('parse'+lineno())
            print('resource: '+str(resource))
            print('vars: '+str(vars(resource)))
        parser = PolicyDocumentParser(debug)
        resource.policy_document = parser.parse(resource.policyDocument)

        return resource



from __future__ import absolute_import, division, print_function
import inspect
import sys

def lineno():
    """Returns the current line number in our program."""
    return str(' - IpAddr - line number: '+str(inspect.currentframe().f_back.f_lineno))


class IpAddr:
    """
    Ip address
    """
    @staticmethod
    def ip4_open(ingress, debug=False):
        """
        Whether ip4 is open to the world
        :param debug:
        :return: boolean
        """
        if debug:
            print('ip4_open'+lineno())
            print("\ningress: "+str(ingress)+lineno())
            if hasattr(ingress,'__dict__'):
                print('vars: '+str(vars(ingress))+lineno())
                print('type:'+str(ingress)+lineno())
                if inspect.isclass(ingress):
                    print(str(vars(ingress)))

        if type(ingress)== type(dict()):
            if debug:
                print("\ningress is a dict"+lineno())

            if type(ingress['cidr_blocks']) == type(list()):
                for block in ingress['cidr_blocks']:
                    if block == '0.0.0.0/0':
                        return True

        elif type(ingress) == type(list()):
            if debug:
                print("ingress is a list"+lineno())
            for item in ingress:

                if debug:
                    print('item is: '+str(item)+lineno())

                if type(ingress['cidr_blocks']) == type(list()):
                    for block in ingress['cidr_blocks']:
                        if block == '0.0.0.0/0':
                            return True

        elif hasattr(ingress, 'cfn_model'):
            if debug:
                print('ingress is object and has cfn_model attribute'+lineno())
                print('model: '+str(ingress.cfn_model)+lineno())

            if hasattr(ingress.cfn_model, 'cfn_model'):
                if debug:
                    print("\ningress cfn_model object has a cfn_model"+lineno())

                if type(ingress.cfn_model.cfn_model) == type(dict()):

                    if ingress.cfn_model.cfn_model['CidrIp'] == '0.0.0.0/0':
                        return True

                elif type(ingress.cfn_model.cfn_model) == type(list()):
                    for item in ingress.cfn_model.cfn_model:
                        if item == 'CidrIp' and item['CidrIp'] == '0.0.0.0/0':
                            return True

            elif hasattr(ingress, 'cidrIp'):
                if debug:
                    print("\ningress had cidrIp attribute"+lineno())
                if type(ingress.cidrIp) == type(str()) and str(ingress.cidrIp) == '0.0.0.0/0':
                    return True

                if sys.version_info[0] < 3:
                    if type(ingress.cidrIp) == type(unicode()) and str(ingress.cidrIp) == '0.0.0.0/0':
                        return True

            else:
                if debug:
                    print("\n"+str(ingress.cfn_model)+lineno())

                if hasattr(ingress,'cidr_blocks'):
                    if debug:
                        print('has cidr blocks '+lineno())

                    for block in ingress.cidr_blocks:
                        if block == '0.0.0.0/0':
                            return True
                else:

                    if hasattr(ingress.cfn_model,'cfn_model'):
                        if debug:
                            print(str(ingress.cfn_model.cfn_model)+lineno())
                        if 'Properties' in ingress.cfn_model.cfn_model:
                            if 'CidrIp' in ingress.cfn_model['Properties']:
                                if type(ingress.cfn_model.cfn_model['Properties']['CidrIp']) == type(str()) and str(
                                        ingress.cfn_model.cfn_model['Properties']['CidrIp']) == '0.0.0.0/0':
                                    return True

                                if sys.version_info[0] < 3:
                                    if type(ingress.cfn_model.cfn_model['Properties']['CidrIp']) == type(unicode()) and str(
                                            ingress.cfn_model.cfn_model['Properties']['CidrIp']) == '0.0.0.0/0':
                                        return True

                    elif type(ingress) == type(dict()):
                        if debug:
                            print('is a dict'+lineno())

                        if ingress['CidrIp'] == '0.0.0.0/0':
                            return True

                    elif type(ingress) == type(list()):

                        if debug:
                            print('is a list'+lineno())

                        for item in ingress:
                            if item == 'CidrIp' and item['CidrIp'] == '0.0.0.0/0':
                                return True

        return False

    @staticmethod
    def ip6_open(ingress, debug=False):
        """
        if ipv6 is open to world
        :param debug:
        :return: boolean
        """
        if debug:
            print('ip6_open'+lineno())
            print('ingress: '+str(ingress)+lineno())
            if hasattr(ingress,'__dict__'):
                print('vars: '+str(vars(ingress))+lineno())


        if type(ingress)==type(dict()):
            if ingress['CidrIp'] == '::/0':
                return True

        elif type(ingress) == type(list()):

            for item in ingress:
                if 'CidrIp' in item and item['CidrIp'] == '::/0':
                    return True

        elif hasattr(ingress,'cfn_model'):
            if debug:
                print('ingress is object and has cfn_model attribute'+lineno())

            if hasattr(ingress.cfn_model, 'cfn_model'):
                if debug:
                    print("\ningress cfn_model object has a cfn_model"+lineno())
                if type(ingress.cfn_model.cfn_model) == type(dict()):

                    if ingress.cfn_model.cfn_model['CidrIp'] == '::/0':
                        return True

                elif type(ingress.cfn_model.cfn_model) == type(list()):

                    for item in ingress.cfn_model.cfn_model:
                        if item == 'CidrIp' and item['CidrIp'] == '::/0':
                            return True
        else:
            if debug:
                print('var:' + str(vars(ingress)) + lineno())

            if hasattr(ingress,'cidrIp'):
                if type(ingress.cidrIp) == type(str()) and str(ingress.cidrIp) == '::/0':
                    return True

                if sys.version_info[0] < 3:
                    if type(ingress.cidrIp) == type(unicode()) and str(ingress.cidrIp) == '::/0':
                        return True

            elif hasattr(ingress,'cidrIpv6'):

                if type(ingress.cidrIpv6) == type(str()) and str(ingress.cidrIpv6) == '::/0':
                    return True

                if sys.version_info[0] < 3:
                    if type(ingress.cidrIpv6) == type(unicode()) and str(ingress.cidrIpv6) == '::/0':
                        return True

            elif hasattr(ingress, 'cfn_model'):

                if 'Properties' in ingress.cfn_model:
                    if 'CidrIp' in ingress.cfn_model['Properties']:
                        if type(ingress.cfn_model['Properties']['CidrIp']) == type(str()) and str(ingress.cfn_model['Properties']['CidrIp']) == '::/0':
                            return True

                        if sys.version_info[0] < 3:
                            if type(ingress.cfn_model['Properties']['CidrIp']) == type(unicode()) and str(ingress.cfn_model['Properties']['CidrIp']) == '::/0':
                                return True

        return False

    @staticmethod
    def ip4_cidr_range(ingress, debug=False):
        """
        IP4 is not /32
        :param debug:
        :return: boolean
        """
        if debug:
            print('ip4_cidr_range '+str(ingress)+lineno())
            print('type: '+str(type(ingress))+lineno())

        suffix = "/32";

        if type(ingress)==type(dict()):

            if debug:
                print('ingress is a dict: '+lineno())

            if 'CidrIp' in ingress:

                if debug:
                    print('CidrIp in ingress '+lineno())
                    print('type: '+str(type(ingress['CidrIp']))+lineno())

                if type(ingress['CidrIp']) == type(str()):

                    if debug:
                        print('ip is: ' + str(ingress['CidrIp']) + lineno())

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress['CidrIp']:
                        return True

                    elif ingress['CidrIp'].endswith(suffix):

                        if debug:
                            print('ip ends with /32' + lineno())

                        return True
                    else:
                        if debug:
                            print('ip does not end with /32'+lineno())
                        return False

                if sys.version_info[0] < 3 and type(ingress['CidrIp']) == type(unicode()):

                    if debug:
                        print('ip is: ' + str(ingress['CidrIp']) + lineno())

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress['CidrIp']:
                        return True

                    elif ingress['CidrIp'].endswith(suffix):

                        if debug:
                            print('ip ends with /32' + lineno())

                        return True
                    else:
                        if debug:
                            print('ip does not end with /32'+lineno())
                        return False

        elif type(ingress) == type(list()):
            if debug:
                print('is a list: '+lineno())

            for item in ingress:
                if 'CidrIp' in item:
                    if type(item['CidrIp']) == type(str()):

                        if debug:
                            print('ip is: ' + str(item['CidrIp']) + lineno())

                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in item['CidrIp']:
                            return True

                        elif item['CidrIp'].endswith(suffix):
                            if debug:
                                print('ip ends with /32' + lineno())
                            return True
                        else:
                            if debug:
                                print('ip does not end with /32'+lineno())
                            return False

                    if sys.version_info[0] < 3 and type(item['CidrIp']) == type(unicode()):

                            if debug:
                                print('ip is: ' + str(item['CidrIp']) + lineno())

                            # only care about literals.  if a Hash/Ref not going to chase it down
                            # given likely a Parameter with external val
                            if 'Ref' in item['CidrIp']:
                                return True

                            elif item ['CidrIp'].endswith(suffix):
                                if debug:
                                    print('ip ends with /32' + lineno())
                                return True
                            else:
                                if debug:
                                    print('ip does not end with /32'+lineno())
                                return False


        elif hasattr(ingress,'cidrIp'):

            if type(ingress.cidrIp) == type(str()):

                if debug:
                    print('ip is: '+str(ingress.cidrIp)+lineno())

                if type(ingress.cidrIp) == type(list()):

                    for item in ingress:
                        if 'CidrIp' in item:
                            if type(item['CidrIp']) == type(str()):

                                if debug:
                                    print('ip is: ' + str(item['CidrIp']) + lineno())

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item['CidrIp']:
                                    return True

                                elif item['CidrIp'].endswith(suffix):

                                    if debug:
                                        print('ip ends with /32' + lineno())

                                    return True

                                else:
                                    if debug:
                                        print('ip does not end with /32' + lineno())
                                    return False

                            if sys.version_info[0] < 3 and type(item['CidrIp']) == type(unicode()):

                                    if debug:
                                        print('ip is: ' + str(item['CidrIp']) + lineno())

                                    # only care about literals.  if a Hash/Ref not going to chase it down
                                    # given likely a Parameter with external val
                                    if 'Ref' in item['CidrIp']:
                                        return True

                                    elif item['CidrIp'].endswith(suffix):

                                        if debug:
                                            print('ip ends with /32' + lineno())

                                        return True

                                    else:
                                        if debug:
                                            print('ip does not end with /32' + lineno())
                                        return False


                elif type(ingress.cidrIp) == type(dict()):

                    for item in ingress.cidrIp:
                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in ingress.cidrIp[item]:
                            return True

                        elif item == 'Ref':
                            return True

                        elif ingress.cidrIp[item].endswith(suffix):
                            if debug:
                                print('ip ends with /32' + lineno())
                            return True

                        else:
                            if debug:
                                print('ip does not end with /32'+lineno())
                            return False

                elif type(ingress.cidrIp) == type(str()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIp:
                        return True

                    elif ingress.cidrIp.endswith(suffix):
                        if debug:
                            print('ip ends with /32' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /32' + lineno())
                        return False

                elif sys.version_info[0] < 3 and  type(ingress.cidrIp) == type(unicode()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIp:
                        return True

                    elif ingress.cidrIp.endswith(suffix):
                        if debug:
                            print('ip ends with /32' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /32' + lineno())
                        return False
                else:
                    print('not sure what this is')
                    print('need to fix')
                    sys.exit(1)

            elif sys.version_info[0] < 3 and type(ingress.cidrIp) == type(unicode()):

                if debug:
                    print('ip is: '+str(ingress.cidrIp)+lineno())

                if type(ingress.cidrIp) == type(list()):

                    for item in ingress:
                        if 'CidrIp' in item:
                            if type(item['CidrIp']) == type(str()):

                                if debug:
                                    print('ip is: ' + str(item['CidrIp']) + lineno())

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item['CidrIp']:
                                    return True

                                elif item['CidrIp'].endswith(suffix):

                                    if debug:
                                        print('ip ends with /32' + lineno())

                                    return True

                                else:
                                    if debug:
                                        print('ip does not end with /32' + lineno())
                                    return False

                            if sys.version_info[0] < 3:
                                if type(item['CidrIp']) == type(unicode()):

                                    if debug:
                                        print('ip is: ' + str(item['CidrIp']) + lineno())

                                    # only care about literals.  if a Hash/Ref not going to chase it down
                                    # given likely a Parameter with external val
                                    if 'Ref' in item['CidrIp']:
                                        return True

                                    elif item['CidrIp'].endswith(suffix):

                                        if debug:
                                            print('ip ends with /32' + lineno())

                                        return True

                                    else:
                                        if debug:
                                            print('ip does not end with /32' + lineno())
                                        return False

                elif type(ingress.cidrIp) == type(dict()):

                    for item in ingress.cidrIp:
                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in ingress.cidrIp[item]:
                            return True

                        elif item == 'Ref':
                            return True

                        elif ingress.cidrIp[item].endswith(suffix):
                            if debug:
                                print('ip ends with /32' + lineno())
                            return True

                        else:
                            if debug:
                                print('ip does not end with /32'+lineno())
                            return False

                elif type(ingress.cidrIp) == type(str()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIp:
                        return True

                    elif ingress.cidrIp.endswith(suffix):
                        if debug:
                            print('ip ends with /32' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /32' + lineno())
                        return False

                elif sys.version_info[0] < 3 and type(ingress.cidrIp) == type(unicode()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIp:
                        return True

                    elif ingress.cidrIp.endswith(suffix):
                        if debug:
                            print('ip ends with /32' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /32' + lineno())
                        return False
                else:
                    print('not sure what this is')
                    print('need to fix')
                    sys.exit(1)

            else:

                if debug:
                    print('ip is: ' + str(ingress.cidrIp) + lineno())
                    print('type: '+str(type(ingress.cidrIp))+lineno())

                if type(ingress.cidrIp) == type(dict()):

                    if debug:
                        print('is a dict: '+lineno())

                    for item in ingress.cidrIp:

                        if debug:
                            print('item: '+str(item)+lineno())

                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in ingress.cidrIp[item]:
                            if debug:
                                print('is a ref - ignoring = '+lineno())
                            return True

                        elif item == 'Ref':
                            if debug:
                                print('is a ref - ignoring - '+lineno())
                            return True

                        elif ingress.cidrIp[item].endswith(suffix):
                            if debug:
                                print('ip ends with /32' + lineno())
                            return True

                        else:
                            if debug:
                                print('ip does not end with /32'+lineno())
                            return False

                elif type(ingress.cidrIp) == type(list()):

                    if debug:
                      print('is a list: '+lineno())

                    has_invalid_cidr = False

                    for item in ingress.cidrIp:

                        if debug:
                            print('list item: '+str(item)+lineno())


                        if type(item) == type(dict()):

                            for item2 in item:

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item[item2]:
                                    return True

                                elif item2 == 'Ref':
                                    return True

                                elif item[item2].endswith(suffix):
                                    if debug:
                                        print('ip ends with /32' + lineno())
                                    return True

                                else:
                                    if debug:
                                        print('ip does not end with /32' + lineno())
                                    return False

                        elif 'CidrIp' in item:
                            if type(item['CidrIp']) == type(str()):

                                if debug:
                                    print('ip is: ' + str(item['CidrIp']) + lineno())

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item['CidrIp']:
                                    has_invalid_cidr= True

                                elif item['CidrIp'].endswith(suffix):
                                    if debug:
                                        print('ip ends with /32' + lineno())
                                    return True
                                else:
                                    if debug:
                                        print('ip does not end with /32' + lineno())
                                    has_invalid_cidr= False

                            if sys.version_info[0] < 3 and type(item['CidrIp']) == type(unicode()):

                                    if debug:
                                        print('ip is: ' + str(item['CidrIp']) + lineno())

                                    # only care about literals.  if a Hash/Ref not going to chase it down
                                    # given likely a Parameter with external val
                                    if 'Ref' in item['CidrIp']:
                                        has_invalid_cidr= True

                                    elif item['CidrIp'].endswith(suffix):
                                        if debug:
                                            print('ip ends with /32' + lineno())
                                        return True
                                    else:
                                        if debug:
                                            print('ip does not end with /32' + lineno())
                                        has_invalid_cidr= False

                        return has_invalid_cidr

                else:
                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIp:
                        return True

                    elif ingress.cidrIp.endswith(suffix):
                        if debug:
                            print('ip ends with /32' + lineno())
                        return True

                    else:
                        if debug:
                            print('ip does not end with /32'+lineno())
                        return False

        elif type(ingress) == type(str()):
            if debug:
                print('is a str '+lineno())


            # only care about literals.  if a Hash/Ref not going to chase it down
            # given likely a Parameter with external val
            if 'Ref' in ingress:
                return True

            elif ingress.endswith('/32'):
                return True
            else:
                if debug:
                    print('ip does not end with /32' + lineno())
                return False
        elif sys.version_info[0] < 3 and type(ingress) == type(unicode()):
            if debug:
                print('is a str '+lineno())


            # only care about literals.  if a Hash/Ref not going to chase it down
            # given likely a Parameter with external val
            if 'Ref' in ingress:
                return True

            elif ingress.endswith('/32'):
                return True
            else:
                if debug:
                    print('ip does not end with /32' + lineno())
                return False
        else:

            print('not sure what type of object this is '+lineno())
            print('vars: '+str(vars(ingress))+lineno())
            sys.exit(1)

        return False

        #ingress.cidrIp.is_a?(String) && !ingress.cidrIp.end_with?('/32')

    @staticmethod
    def ip6_cidr_range(ingress, debug=False):
        """
        If ipv6 is not /128
        :param debug:
        :return: boolean
        """
        if debug:
              print('ip6_cidr_range ' + str(ingress) + lineno())
              print('type: ' + str(type(ingress)) + lineno())
              if hasattr(ingress, '__dict__'):
                  print('vars: ' + str(vars(ingress)) + lineno())

        suffix = "/128";

        if type(ingress) == type(dict()):

            if debug:
                print('ingress is a dict: ' + lineno())

            if 'CidrIp' in ingress:

                if debug:
                    print('CiderIp in ingress '+lineno())

                if type(ingress['CidrIp']) == type(str()):

                    if debug:
                        print('ip is: ' + str(ingress['CidrIp']) + lineno())

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress['CidrIp']:
                        return True

                    elif ingress['CidrIp'].endswith(suffix):
                        if debug:
                            print('ip ends with /128' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /128' + lineno())
                        return False

                elif sys.version_info[0] < 3 and  type(ingress['CidrIp']) == type(unicode()):

                        if debug:
                            print('ip is: ' + str(ingress['CidrIp']) + lineno())

                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in ingress['CidrIp']:
                            return True

                        elif ingress['CidrIp'].endswith(suffix):
                            if debug:
                                print('ip ends with /128' + lineno())
                            return True
                        else:
                            if debug:
                                print('ip does not end with /128' + lineno())
                            return False

        elif type(ingress) == type(list()):

            for item in ingress:
                if 'CidrIp' in item:
                    if type(item['CidrIp']) == type(str()):

                        if debug:
                            print('ip is: ' + str(item['CidrIp']) + lineno())

                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in item['CidrIp']:
                            return True

                        elif item['CidrIp'].endswith(suffix):
                            if debug:
                                print('ip ends with /128' + lineno())
                            return True
                        else:
                            if debug:
                                print('ip does not end with /128' + lineno())
                            return False

                    if sys.version_info[0] < 3 and  type(item['CidrIp']) == type(unicode()):

                            if debug:
                                print('ip is: ' + str(item['CidrIp']) + lineno())

                            # only care about literals.  if a Hash/Ref not going to chase it down
                            # given likely a Parameter with external val
                            if 'Ref' in item['CidrIp']:
                                return True

                            elif item['CidrIp'].endswith(suffix):
                                if debug:
                                    print('ip ends with /128' + lineno())
                                return True
                            else:
                                if debug:
                                    print('ip does not end with /128' + lineno())
                                return False

        elif hasattr(ingress, 'cidrIpv6'):

            if type(ingress.cidrIpv6) == type(str()):

                if debug:
                    print('ip is: ' + str(ingress.cidrIpv6) + lineno())

                if type(ingress.cidrIpv6) == type(list()):

                    for item in ingress:
                        if 'CidrIp' in item:
                            if type(item['CidrIp']) == type(str()):

                                if debug:
                                    print('ip is: ' + str(item['CidrIp']) + lineno())

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item['CidrIp']:
                                    return True

                                elif item['CidrIp'].endswith(suffix):
                                    if debug:
                                        print('ip ends with /128' + lineno())
                                    return True
                                else:
                                    if debug:
                                        print('ip does not end with /128' + lineno())
                                    return False
                            if sys.version_info[0] < 3:
                                if type(item['CidrIp']) == type(unicode()):

                                    if debug:
                                        print('ip is: ' + str(item['CidrIp']) + lineno())

                                    # only care about literals.  if a Hash/Ref not going to chase it down
                                    # given likely a Parameter with external val
                                    if 'Ref' in item['CidrIp']:
                                        return True

                                    elif item['CidrIp'].endswith(suffix):
                                        if debug:
                                            print('ip ends with /128' + lineno())
                                        return True
                                    else:
                                        if debug:
                                            print('ip does not end with /128' + lineno())
                                        return False

                elif type(ingress.cidrIpv6) == type(dict()):

                    for item in ingress.cidrIp:
                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in ingress.cidrIpv6[item]:
                            return True

                        elif ingress.cidrIpv6[item].endswith(suffix):
                            if debug:
                                print('ip ends with /128' + lineno())
                            return True

                        else:
                            if debug:
                                print('ip does not end with /128' + lineno())
                            return False
                elif type(ingress.cidrIpv6) == type(str()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIpv6:
                        return False

                    elif ingress.cidrIpv6.endswith(suffix):
                        if debug:
                            print('ip ends with /128' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /128' + lineno())
                        return False

                elif sys.version_info[0] < 3 and type(ingress.cidrIpv6) == type(unicode()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIpv6:
                        return False

                    elif ingress.cidrIpv6.endswith(suffix):
                        if debug:
                            print('ip ends with /128' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /128' + lineno())
                        return False
                else:
                    print('not sure what this is')
                    print('need to fix')
                    sys.exit(1)

            elif sys.version_info[0] < 3 and type(ingress.cidrIpv6) == type(unicode()):

                if debug:
                    print('ip is: ' + str(ingress.cidrIpv6) + lineno())

                if type(ingress.cidrIpv6) == type(list()):

                    for item in ingress:
                        if 'CidrIp' in item:
                            if type(item['CidrIp']) == type(str()):

                                if debug:
                                    print('ip is: ' + str(item['CidrIp']) + lineno())

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item['CidrIp']:
                                    return True

                                elif item['CidrIp'].endswith(suffix):
                                    if debug:
                                        print('ip ends with /128' + lineno())
                                    return True
                                else:
                                    if debug:
                                        print('ip does not end with /128' + lineno())
                                    return False
                            if sys.version_info[0] < 3:
                                if type(item['CidrIp']) == type(unicode()):

                                    if debug:
                                        print('ip is: ' + str(item['CidrIp']) + lineno())

                                    # only care about literals.  if a Hash/Ref not going to chase it down
                                    # given likely a Parameter with external val
                                    if 'Ref' in item['CidrIp']:
                                        return True

                                    elif item['CidrIp'].endswith(suffix):
                                        if debug:
                                            print('ip ends with /128' + lineno())
                                        return True
                                    else:
                                        if debug:
                                            print('ip does not end with /128' + lineno())
                                        return False

                elif type(ingress.cidrIpv6) == type(dict()):

                    for item in ingress.cidrIp:
                        # only care about literals.  if a Hash/Ref not going to chase it down
                        # given likely a Parameter with external val
                        if 'Ref' in ingress.cidrIpv6[item]:
                            return True

                        elif ingress.cidrIpv6[item].endswith(suffix):
                            if debug:
                                print('ip ends with /128' + lineno())
                            return True

                        else:
                            if debug:
                                print('ip does not end with /128' + lineno())
                            return False
                elif type(ingress.cidrIpv6) == type(str()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIpv6:
                        return False

                    elif ingress.cidrIpv6.endswith(suffix):
                        if debug:
                            print('ip ends with /128' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /128' + lineno())
                        return False

                elif sys.version_info[0] < 3 and type(ingress.cidrIpv6) == type(unicode()):

                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIpv6:
                        return False

                    elif ingress.cidrIpv6.endswith(suffix):
                        if debug:
                            print('ip ends with /128' + lineno())
                        return True
                    else:
                        if debug:
                            print('ip does not end with /128' + lineno())
                        return False
                else:
                    print('not sure what this is')
                    print('need to fix')
                    sys.exit(1)

            else:
                if debug:
                    print('ip is: ' + str(ingress.cidrIpv6) + lineno())
                    print('type: ' + str(type(ingress.cidrIpv6)) + lineno())

                if type(ingress.cidrIpv6) == type(list()):

                    has_invalid_cidr = False

                    for item in ingress.cidrIpv6:

                        if debug:
                            print('list item: ' + str(item) + lineno())

                        if type(item) == type(dict()):

                            for item2 in item:
                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item[item2]:
                                    return True

                                elif item2 == 'Ref':
                                    return True

                                elif item[item2].endswith(suffix):
                                    if debug:
                                        print('ip ends with /32' + lineno())
                                    return True

                                else:
                                    if debug:
                                        print('ip does not end with /32' + lineno())
                                    return False

                        elif 'CidrIp' in item:
                            if type(item['CidrIp']) == type(str()):

                                if debug:
                                    print('ip is: ' + str(item['CidrIp']) + lineno())

                                # only care about literals.  if a Hash/Ref not going to chase it down
                                # given likely a Parameter with external val
                                if 'Ref' in item['CidrIp']:
                                    has_invalid_cidr = True

                                elif item['CidrIp'].endswith(suffix):
                                    if debug:
                                        print('ip ends with /128' + lineno())
                                    return True
                                else:
                                    if debug:
                                        print('ip does not end with /128' + lineno())
                                    has_invalid_cidr = False

                            if sys.version_info[0] < 3:
                                if type(item['CidrIp']) == type(unicode()):

                                    if debug:
                                        print('ip is: ' + str(item['CidrIp']) + lineno())

                                    # only care about literals.  if a Hash/Ref not going to chase it down
                                    # given likely a Parameter with external val
                                    if 'Ref' in item['CidrIp']:
                                        has_invalid_cidr = True

                                    elif item['CidrIp'].endswith(suffix):
                                        if debug:
                                            print('ip ends with /128' + lineno())
                                        return True
                                    else:
                                        if debug:
                                            print('ip does not end with /128' + lineno())
                                        has_invalid_cidr = False

                        return has_invalid_cidr

                else:
                    # only care about literals.  if a Hash/Ref not going to chase it down
                    # given likely a Parameter with external val
                    if 'Ref' in ingress.cidrIpv6:
                        return True

                    elif ingress.cidrIpv6.endswith(suffix):
                        if debug:
                            print('ip ends with /128' + lineno())
                        return True

                    else:
                        if debug:
                            print('ip does not end with /128' + lineno())
                        return False

        elif type(ingress) == type(str()):
            if debug:
                print('is a str ' + lineno())

            # only care about literals.  if a Hash/Ref not going to chase it down
            # given likely a Parameter with external val
            if 'Ref' in ingress:
                return True

            elif ingress.endswith('/128'):
                return True
            else:
                if debug:
                    print('ip does not end with /128' + lineno())
                return False
        elif sys.version_info[0] < 3 and type(ingress) == type(unicode()):
            if debug:
                print('is a str ' + lineno())

            # only care about literals.  if a Hash/Ref not going to chase it down
            # given likely a Parameter with external val
            if 'Ref' in ingress:
                return True

            elif ingress.endswith('/128'):
                return True
            else:
                if debug:
                    print('ip does not end with /128' + lineno())
                return False

        return False


from __future__ import absolute_import, division, print_function
import sys
import inspect
import json
import re
from terraform_model.model.PolicyDocument import PolicyDocument
from terraform_model.model.Statement import Statement


def lineno():
    """Returns the current line number in our program."""
    return str(' - PolicyDocumentParser - line number: '+str(inspect.currentframe().f_back.f_lineno))


class PolicyDocumentParser:
    """
    Policy document parser
    """

    def __init__(self, debug=False):
        """
        Initialize
        :param debug: 
        """
        self.debug = debug

        if self.debug:
            print('ParserError - init'+lineno())

    def parse(self, cfn_model, raw_policy_document):

        if self.debug:
            print("\n\n################################")
            print('PolicyDocumentParser - parse'+lineno())
            print("####################################\n\n")

            print('raw_policy_document: '+str(raw_policy_document)+lineno())

        policy_document = PolicyDocument(debug=self.debug)

        if self.debug:
            print('have policy document: '+lineno())

        if type(policy_document) == type(str()):
            if self.debug:
                print('policy document is a string: '+lineno())

            json_acceptable_string = policy_document.replace("'", "\"")
            policy_document= json.loads(json_acceptable_string)


        if type(raw_policy_document) == type(str()):
            if self.debug:
                print('raw policy document is a string: '+lineno())

            matchObj = re.match(r'(.*)(\${)([^}]+)(})(.*)', raw_policy_document, re.M | re.I)

            if matchObj:
                prefix = str(matchObj.group(1))
                lookup = str(matchObj.group(3))
                suffix=str(matchObj.group(5))

                if self.debug:
                    print('prefix: '+str(prefix)+lineno())
                    print('lookup: '+str(lookup)+lineno())
                    print('suffix: '+str(suffix)+lineno())

                parts = lookup.split('.')

                if self.debug:
                    print('parts: '+str(parts)+lineno())
                    print('cfn model - raw model: '+str(cfn_model.raw_model)+lineno())

                for i in range(len(parts)):

                    if parts[i] in cfn_model.raw_model:
                        if self.debug:
                            print('part '+str(i)+' '+str(parts[i])+' in cfn model')

                        if parts[i+1] in cfn_model.raw_model[parts[i]]:
                            if self.debug:
                                print('part ' + str(i+1) + ' ' + str(parts[i+1]) + ' in cfn model')

                            if parts[i+2] in cfn_model.raw_model[parts[i]][parts[i+1]]:
                                if self.debug:
                                    print('part ' + str(i + 2) + ' ' + str(parts[i + 2]) + ' in cfn model')

                                raw_policy_document= str(prefix)+str(cfn_model.raw_model[parts[i]][parts[i+1]][parts[i+2]])+str(suffix)

                                if self.debug:
                                    print('new raw policy is: '+str(raw_policy_document)+lineno())

                                break



            json_acceptable_string = raw_policy_document.replace("'", "\"")
            raw_policy_document= json.loads(json_acceptable_string)

        if 'Version' in raw_policy_document:
            if self.debug:
                print('version in policy document: '+lineno())
                print('type: '+str(type(raw_policy_document)))

                print('version: '+str(raw_policy_document['Version'])+lineno())

            policy_document.version = raw_policy_document['Version']

        if 'Statement' in raw_policy_document:
            if self.debug:
                print('has statement in raw policy document'+lineno())
                print('statement: '+str(raw_policy_document['Statement']))

            if type(raw_policy_document['Statement'])==type(list()):

                if len(raw_policy_document['Statement'])>1:
                    if self.debug:
                        print('more than one statemnt')

                for statement in raw_policy_document['Statement']:
                    if self.debug:
                        print('statement: '+str(statement)+lineno())
                    streamlined_array = self.streamline_array(statement)
                    parsed_statement = self.parse_statement(streamlined_array)
                    if self.debug:
                        print('parsed_statement: '+str(parsed_statement)+lineno())
                    policy_document.statements.append(parsed_statement)

            elif type(raw_policy_document['Statement'])== type(dict()):

                if self.debug:
                    print('statement: ' + str(raw_policy_document['Statement'])+lineno())

                parsed_statement = self.parse_statement(raw_policy_document['Statement'])
                if self.debug:
                    print('parsed_statement: ' + str(parsed_statement) + lineno())
                policy_document.statements.append(parsed_statement)


            elif type(raw_policy_document['Statement']) == type(str()):
                if self.debug:
                    print('statement: ' + str(raw_policy_document['Statement'])+lineno())

                parsed_statement = self.parse_statement(raw_policy_document['Statement'])
                if self.debug:
                    print('parsed_statement: ' + str(parsed_statement) + lineno())
                policy_document.statements.append(parsed_statement)


            else:
                if self.debug:
                    print('unknown type: '+lineno())
                sys.exit(1)

            if self.debug:
                print('policy_document: '+str(vars(policy_document))+lineno())
                print('statments'+str(policy_document.statements)+lineno())
                print('statement types: '+str(type(policy_document.statements))+lineno())
                print('type: '+str(type(policy_document))+lineno())

        return policy_document



    def parse_statement(self, raw_statement):
        """
        Parse statement
        :param vraw_statement: 
        :return: 
        """
        if self.debug:
            print('parse_statement'+lineno())
            print('raw_statement: '+str(raw_statement)+lineno())
            print('raw statement type is: '+str(type(raw_statement))+lineno())


        statement = Statement(self.debug)
        if 'Effect' in raw_statement:
            statement.effect = raw_statement['Effect']
        if 'Sid' in raw_statement:
            statement.sid = raw_statement['Sid']
        if 'Condition' in raw_statement:
            statement.condition = raw_statement['Condition']
        if 'Action' in raw_statement:
            statement.actions.append(raw_statement['Action'])
        if 'NotAction' in raw_statement:
            statement.not_actions.append(raw_statement['NotAction'])
        if 'Resource' in raw_statement:
            statement.resources.append(raw_statement['Resource'])
        if 'NotResource' in raw_statement:
            statement.not_resources.append(raw_statement['NotResource'])
        if 'Principal' in raw_statement:
            statement.principal = raw_statement['Principal']
        if 'NotPrincipal' in raw_statement:
            statement.not_principal = raw_statement['NotPrincipal']

        if self.debug:
            print('raw_statement: '+str(raw_statement)+lineno())
            if statement.condition:
                print('condition: '+str(statement.condition)+lineno())
            if statement.actions:
                print('actions: '+str(statement.actions)+lineno())
            if statement.sid:
                print('sid: '+str(statement.sid)+lineno())
            if statement.effect:
                print('effect: '+str(statement.effect)+lineno())
            if statement.resources:
                print('resources: '+str(statement.resources)+lineno())
            if statement.principal:
                print('principal: '+str(statement.principal)+lineno())
            if statement.not_principal:
                print('not_principal: '+str(statement.not_principal)+lineno())

            print("\n\n###############################################")
            print("statement: "+str(vars(statement))+lineno())
            print(' wildcard_actions:'+str(statement. wildcard_actions())+lineno())
            print('wildcard_principal: '+str(statement.wildcard_principal())+lineno())
            print('wildcard_resources: '+str(statement.wildcard_resources())+lineno())
            print("#####################################################\n")

        return statement

    def streamline_array(self, one_or_more):
        """
        ???
        :param one_or_more: 
        :return: 
        """

        if one_or_more:
            if self.debug:
                print('one_or_more: '+str(one_or_more)+lineno())


            if type(one_or_more) == type(str()):
                if self.debug:
                    print('is a string '+lineno())
                return one_or_more
            elif type(one_or_more)==type(dict()):
                if self.debug:
                    print('is a dict: '+lineno())


                return one_or_more
            elif type(one_or_more)==type(list()):
                if self.debug:
                    print('is an array: '+lineno())
                return one_or_more

            else:
                raise "unexpected object in streamline_array "+str(one_or_more)
        # FIXME
        sys.exit(1)
        return None


